# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------


"""Tools for analysing the service for deployment readiness"""

import copy
import logging
import os
import re
from typing import Any, Dict, List, Optional, OrderedDict, Set, Union, cast

from aea.configurations.base import AgentConfig, SkillConfig
from aea.configurations.data_types import (
    ComponentId,
    ComponentType,
    PackageId,
    PublicId,
)
from aea.crypto.base import LedgerApi
from aea.helpers.cid import to_v0
from aea.helpers.env_vars import ENV_VARIABLE_RE, apply_env_variables
from aea.helpers.logging import setup_logger
from jsonschema.exceptions import ValidationError as SchemaValidationError
from jsonschema.validators import Draft4Validator
from requests.exceptions import ConnectionError as RequestConnectionError

from autonomy.chain.base import ServiceState
from autonomy.chain.config import ChainType
from autonomy.chain.service import get_service_info
from autonomy.configurations.base import Service


ABCI_CONNECTION_SCHEMA = {
    "type": "object",
    "properties": {
        "config": {
            "type": "object",
            "required": [
                "host",
                "port",
                "use_tendermint",
            ],
        },
        "required": ["config"],
    },
}

LEDGER_CONNECTION_SCHEMA = {
    "type": "object",
    "properties": {
        "config": {
            "type": "object",
            "properties": {
                "ledger_apis": {
                    "type": "object",
                    "properties": {
                        "ethereum": {
                            "type": "object",
                            "required": [
                                "address",
                                "chain_id",
                                "poa_chain",
                                "default_gas_price_strategy",
                            ],
                        },
                    },
                    "additionalProperties": False,
                    "minProperties": 1,
                }
            },
            "required": ["ledger_apis"],
        },
    },
    "required": ["config"],
}

ABCI_SKILL_SCHEMA = {
    "type": "object",
    "properties": {
        "models": {
            "type": "object",
            "properties": {
                "params": {
                    "type": "object",
                    "properties": {
                        "args": {
                            "type": "object",
                            "properties": {
                                "setup": {
                                    "type": "object",
                                    "required": [
                                        "safe_contract_address",
                                        "all_participants",
                                        "consensus_threshold",
                                    ],
                                }
                            },
                            "required": [
                                "setup",
                                "service_registry_address",
                                "share_tm_config_on_startup",
                                "on_chain_service_id",
                            ],
                        },
                    },
                    "required": [
                        "args",
                    ],
                },
            },
            "required": [
                "params",
            ],
        }
    },
    "required": ["models"],
}

ABCI_SKILL_MODEL_PARAMS_SCHEMA = {
    "type": "object",
    "properties": {
        "setup": {
            "type": "object",
            "required": [
                "safe_contract_address",
                "all_participants",
                "consensus_threshold",
            ],
        }
    },
    "required": [
        "genesis_config",
        "service_id",
        "max_healthcheck",
        "round_timeout_seconds",
        "sleep_time",
        "retry_timeout",
        "retry_attempts",
        "keeper_timeout",
        "reset_pause_duration",
        "drand_public_key",
        "tendermint_url",
        "tendermint_com_url",
        "tendermint_p2p_url",
        "tendermint_max_retries",
        "tendermint_check_sleep_delay",
        "reset_tendermint_after",
        "cleanup_history_depth",
        "cleanup_history_depth_current",
        "request_timeout",
        "request_retry_delay",
        "tx_timeout",
        "max_attempts",
        "service_registry_address",
        "on_chain_service_id",
        "share_tm_config_on_startup",
        "use_termination",
        "setup",
    ],
}

ABCI = "abci"
LEDGER = "ledger"

REQUIRED_PROPERTY_RE = re.compile(r"'(.*)' is a required property")
PROPERTY_NOT_ALLOWED_RE = re.compile(
    r"Additional properties are not allowed \((('(.*)'(,)? )+)(was|were) unexpected\)"
)


class CustomSchemaValidationError(SchemaValidationError):
    """Custom schema validation error to report all errors at once."""

    def __init__(
        self,
        extra_properties: Optional[List[str]] = None,
        missing_properties: Optional[List[str]] = None,
        not_having_enough_properties: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize object."""
        self.extra_properties = extra_properties or []
        self.missing_properties = missing_properties or []
        self.not_having_enough_properties = not_having_enough_properties or []
        self.should_raise = False
        error_strings = []
        if len(self.extra_properties) > 0:
            self.should_raise = True
            self.extra_properties_error = "Additional properties found " + ", ".join(
                map(lambda x: f"'{x}'", self.extra_properties)
            )
            error_strings.append(self.extra_properties_error)

        if len(self.missing_properties) > 0:
            self.should_raise = True
            self.missing_properties_error = (
                "Following properties are require but missing "
                + ", ".join(map(lambda x: f"'{x}'", self.missing_properties))
            )
            error_strings.append(self.missing_properties_error)

        if len(self.not_having_enough_properties) > 0:
            self.should_raise = True
            self.not_having_enough_properties_error = ";".join(
                self.not_having_enough_properties
            )
            error_strings.append(self.not_having_enough_properties_error)

        message = ";".join(error_strings)
        super().__init__(message, **kwargs)


class CustomSchemaValidator(Draft4Validator):
    """Custom schema validator to report all missing fields at once."""

    def validate(self, *args: Any, **kwargs: Any) -> None:
        """Validate and raise all errors at once."""
        missing = []
        extra = []
        not_enough_properties = []
        for error in self.iter_errors(*args, **kwargs):
            message = cast(SchemaValidationError, error).message
            missing_property_check = REQUIRED_PROPERTY_RE.match(message)
            if missing_property_check is not None:
                (missing_property,) = cast(re.Match, missing_property_check).groups()
                missing.append(missing_property)
                continue

            property_not_allowed_check = PROPERTY_NOT_ALLOWED_RE.match(message)
            if property_not_allowed_check is not None:
                extra += re.findall(
                    "[a-zA-Z0-9]+", property_not_allowed_check.groups()[0]
                )
                continue

            if "does not have enough properties" in message:
                not_enough_properties.append(message)

        error = CustomSchemaValidationError(
            extra_properties=extra,
            missing_properties=missing,
            not_having_enough_properties=not_enough_properties,
        )
        if error.should_raise:
            raise error


ABCI_CONNECTION_VALIDATOR = CustomSchemaValidator(schema=ABCI_CONNECTION_SCHEMA)
LEDGER_CONNECTION_VALIDATOR = CustomSchemaValidator(schema=LEDGER_CONNECTION_SCHEMA)
ABCI_SKILL_VALIDATOR = CustomSchemaValidator(schema=ABCI_SKILL_SCHEMA)
ABCI_SKILL_MODEL_PARAMS_VALIDATOR = CustomSchemaValidator(
    schema=ABCI_SKILL_MODEL_PARAMS_SCHEMA
)


class ServiceValidationFailed(Exception):
    """Raise when service validation fails."""


class ServiceAnalyser:
    """Tools to analyse a service"""

    def __init__(
        self,
        service_config: Service,
        abci_skill_id: PublicId,
        is_on_chain_check: bool = False,
        logger: Optional[logging.Logger] = None,
        skip_warnings: bool = False,
    ) -> None:
        """Initialise object."""

        self.service_config = service_config
        self.abci_skill_id = abci_skill_id
        self.is_on_chain_check = is_on_chain_check
        self.logger = logger or setup_logger(name="autonomy.analyse")
        self.skip_warning = skip_warnings

    def _warn(self, msg: str) -> None:
        """Raise warning."""
        if self.skip_warning:  # pragma: nocover
            return
        self.logger.warning(msg=msg)

    def check_on_chain_state(
        self,
        ledger_api: LedgerApi,
        chain_type: ChainType,
        token_id: int,
    ) -> None:
        """Check on-chain state of a service."""
        if not self.is_on_chain_check:
            return

        self.logger.info("Checking if the service is deployed on-chain")
        try:
            (*_, _service_state, _) = get_service_info(
                ledger_api=ledger_api, chain_type=chain_type, token_id=token_id
            )
        except RequestConnectionError as e:
            raise ServiceValidationFailed(
                "On chain service validation failed, could not connect to the RPC"
            ) from e

        service_state = ServiceState(_service_state)
        if service_state != ServiceState.DEPLOYED:
            self._warn(
                f"Service needs to be in deployed state on-chain; Current state={service_state}"
            )

    def check_agent_dependencies_published(
        self, agent_config: AgentConfig, ipfs_pins: Set[str]
    ) -> None:
        """Check if the agent package is published or not"""

        if not self.is_on_chain_check:
            return

        self.logger.info("Checking if agent dependencies are published")
        for dependency in agent_config.package_dependencies:
            if to_v0(dependency.package_hash) not in ipfs_pins:
                raise ServiceValidationFailed(
                    f"Package required for service {self.service_config.public_id} is not published on the IPFS registry"
                    f"\n\tagent: {self.service_config.agent.without_hash()}"
                    f"\n\tpackage: {dependency}"
                )
            self.logger.info(
                f"\t{dependency.public_id.without_hash()} of type {dependency.package_type} is present"
            )

    def _check_overrides_recursively(
        self,
        check_from: Dict,
        check_with: Dict,
        check_from_name: str,
        check_with_name: str,
        path: Optional[List[str]] = None,
    ) -> None:
        """
        Check overrides recursively

        :param check_from: Configuration to check from
        :param check_with: Configuration to compare against
        :param check_from_name: Name for the `check_from` config
        :param check_with_name: Name for the `check_with` config
        :param path: JSON path to the object
        """

        path = path or []
        path_str = ".".join(path)

        # service has them and agent does not - warning
        check_from_set, check_with_set = set(check_from), set(check_with)
        missing = check_with_set - check_from_set
        if len(missing) > 0:
            message = (
                f"{check_with_name} contains configuration which is missing from {check_from_name}\n"
                f"\tPath: {path_str}\n"
                f"\tMissing parameters: {missing}\n"
            )
            self._warn(message)

        missing = check_from_set - check_with_set
        if len(missing) > 0:
            message = (
                f"{check_from_name} contains configuration which is missing from {check_with_name}\n"
                f"\tPath: {path_str}\n"
                f"\tMissing parameters: {missing}\n"
            )
            self._warn(message)

        for key in check_with_set.intersection(check_from_set):
            if not isinstance(check_from[key], (dict, Dict, OrderedDict)):
                continue

            self._check_overrides_recursively(
                check_from=check_from[key],
                check_with=check_with[key],
                check_from_name=check_from_name,
                check_with_name=check_with_name,
                path=[*path, key],
            )

    def cross_verify_overrides(  # pylint: disable=too-many-locals
        self,
        agent_config: AgentConfig,
        skill_config: SkillConfig,
    ) -> None:
        """Cross verify overrides between service config and agent config"""

        self.logger.info("Cross verifying overrides between agent and service")
        agent_overrides = set(agent_config.component_configurations)
        service_overrides = {
            PackageId(
                package_type=component_override["type"],
                public_id=PublicId.from_str(component_override["public_id"]),
            )
            for component_override in self.service_config.overrides
        }

        missing_from_service = agent_overrides - service_overrides
        if len(missing_from_service) > 0:
            self._warn(
                "\n\t- ".join(
                    map(
                        str,
                        [
                            "Overrides with following component IDs are defined in the agent configuration but not in the service configuration",
                            *missing_from_service,
                        ],
                    )
                )
            )

        missing_from_agent = service_overrides - agent_overrides
        if len(missing_from_agent) > 0:
            raise ServiceValidationFailed(
                "\n\t- ".join(
                    map(
                        str,
                        [
                            "Overrides with following component IDs are defined in the service configuration but not in the agent configuration",
                            *missing_from_agent,
                        ],
                    )
                )
            )

        skill_override_from_agent = apply_env_variables(
            agent_config.component_configurations[skill_config.package_id],
            env_variables=os.environ.copy(),
        )
        skill_config_to_check: Dict[str, Dict] = {"models": {}}
        skill_config_json = copy.deepcopy(skill_config.json)
        skill_config_to_check["models"]["params"] = {
            "args": skill_config_json["models"]["params"]["args"]
        }
        if "benchmark_tool" in skill_config_json["models"]:
            skill_config_to_check["models"]["benchmark_tool"] = {
                "args": skill_config_json["models"]["benchmark_tool"]["args"]
            }

        for override in self.service_config.overrides:
            (
                skill_override_from_service,
                component_id,
                has_multiple_overrides,
            ) = Service.process_metadata(copy.deepcopy(override))
            if skill_config.component_id == component_id:
                if not has_multiple_overrides:
                    self._check_overrides_recursively(
                        check_from=skill_override_from_service,
                        check_with=skill_override_from_agent,
                        check_from_name=str(self.service_config.package_id),
                        check_with_name=str(agent_config.package_id),
                    )
                    continue

                for (
                    key,
                    _skill_override_from_service,
                ) in skill_override_from_service.items():
                    self.logger.info(f"Verifying the skill override with index {key}")
                    self._check_overrides_recursively(
                        check_from=_skill_override_from_service,
                        check_with=skill_override_from_agent,
                        check_from_name=str(self.service_config.package_id),
                        check_with_name=str(agent_config.package_id),
                    )
                break

        self.logger.info("Cross verifying overrides between skill and agent")
        self._check_overrides_recursively(
            check_from=skill_override_from_agent,
            check_with=skill_config_to_check,
            check_from_name=str(agent_config.package_id),
            check_with_name=str(skill_config.package_id),
        )

    @classmethod
    def validate_override_env_vars(
        cls,
        overrides: Union[OrderedDict, dict],
        validate_env_var_name: bool = False,
        json_path: Optional[List[str]] = None,
    ) -> List[str]:
        """Validate environment variables."""
        errors = []
        json_path = json_path or []
        for key, value in overrides.items():
            if key in (
                "target_skill_id",
                "identifier",
                "ledger_id",
                "message_format",
                "not_after",
                "not_before",
                "save_path",
                "is_abstract",
            ):
                continue  # pragma: nocover
            _json_path = [*json_path, str(key)]
            if isinstance(value, (dict, OrderedDict)):
                errors += cls.validate_override_env_vars(
                    value,
                    validate_env_var_name=validate_env_var_name,
                    json_path=_json_path,
                )
            elif isinstance(value, list):
                for i, obj in enumerate(value):
                    if not isinstance(obj, (dict, OrderedDict)):
                        continue  # pragma: nocover
                    errors += cls.validate_override_env_vars(
                        obj,
                        validate_env_var_name=validate_env_var_name,
                        json_path=[*_json_path, str(i)],
                    )
            else:
                json_path_str = ".".join(_json_path)
                try:
                    re_result = ENV_VARIABLE_RE.match(value)
                    if re_result is None:
                        errors.append(
                            f"`{json_path_str}` needs to be defined as a environment variable"
                        )
                        continue
                    if validate_env_var_name:
                        _, env_var_name, *_ = re_result.groups()
                        if env_var_name is None:
                            errors.append(
                                f"`{json_path_str}` needs environment variable defined in following format "
                                "${ENV_VAR_NAME:DATA_TYPE:DEFAULT_VALUE}"
                            )
                    apply_env_variables(
                        data={key: value}, env_variables=os.environ.copy()
                    )
                except TypeError:
                    errors.append(
                        f"`{json_path_str}` needs to be defined as a environment variable"
                    )
                except (ValueError, KeyError) as e:
                    errors.append(
                        f"`{json_path_str}` validation failed with following error; {e}"
                    )

        return errors

    def validate_agent_override_env_vars(self, agent_config: AgentConfig) -> None:
        """Check if all of the overrides are defined as a env vars in the agent config"""

        self.logger.info(
            f"Validating agent {agent_config.public_id} overrides for environment variable definitions"
        )
        for package_id, override in agent_config.component_configurations.items():
            errors = self.validate_override_env_vars(overrides=override)
            if len(errors) > 0:
                raise ServiceValidationFailed(
                    "\n\t- ".join(
                        [
                            f"{package_id} envrionment variable validation failed with following error",
                            *errors,
                        ]
                    ),
                )

    def validate_service_override_env_vars(self) -> None:
        """Check if all of the overrides are defined as a env vars in the agent config"""
        self.logger.info(
            f"Validating service {self.service_config.public_id} overrides for environment variable definitions"
        )
        errors = []
        for _component_config in self.service_config.overrides:
            (
                component_config,
                component_id,
                _,
            ) = Service.process_metadata(configuration=_component_config.copy())
            errors = self.validate_override_env_vars(
                overrides=component_config, validate_env_var_name=True
            )
            if len(errors) > 0:
                raise ServiceValidationFailed(
                    "\n\t- ".join(
                        [
                            f"{component_id} envrionment variable validation failed with following error",
                            *errors,
                        ]
                    ),
                )

    @staticmethod
    def _validate_override(
        validator: Draft4Validator,
        overrides: Dict,
        has_multiple_overrides: bool,
        error_message: str,
        raise_custom_exception: bool = True,
    ) -> None:
        """Run validator on the given override."""

        def _validate_override(
            validator: Draft4Validator,
            overrides: Dict,
            error_message: str,
        ) -> None:
            try:
                validator.validate(overrides)
            except CustomSchemaValidationError as e:
                if raise_custom_exception:
                    raise ServiceValidationFailed(
                        error_message.format(error=e.message)
                    ) from e
                raise

        if has_multiple_overrides:
            for idx, override in overrides.items():
                _validate_override(
                    validator=validator,
                    overrides=override,
                    error_message=f"{error_message} at override index {idx}",
                )
        else:
            _validate_override(
                validator=validator,
                overrides=overrides,
                error_message=error_message,
            )

    def validate_override(
        self,
        component_id: ComponentId,
        override: Dict,
        has_multiple_overrides: bool,
    ) -> None:
        """Validate override"""

        if component_id.public_id == self.abci_skill_id:
            self._validate_override(
                validator=ABCI_SKILL_VALIDATOR,
                overrides=override,
                has_multiple_overrides=has_multiple_overrides,
                error_message=(
                    "ABCI Skill `"
                    + str(component_id.public_id)
                    + "` override validation failed; {error}"
                ),
            )

        if (
            component_id.component_type == ComponentType.CONNECTION
            and component_id.name == ABCI
        ):
            self._validate_override(
                validator=ABCI_CONNECTION_VALIDATOR,
                overrides=override,
                has_multiple_overrides=has_multiple_overrides,
                error_message="ABCI connection validation failed; {error}",
            )

        if (
            component_id.component_type == ComponentType.CONNECTION
            and component_id.name == LEDGER
        ):
            try:
                self._validate_override(
                    validator=LEDGER_CONNECTION_VALIDATOR,
                    overrides=override,
                    has_multiple_overrides=has_multiple_overrides,
                    error_message="{error}",
                    raise_custom_exception=False,
                )
            except CustomSchemaValidationError as e:
                if len(e.not_having_enough_properties) > 0:
                    raise ServiceValidationFailed(
                        f"{component_id} override needs at least one ledger API definition"
                    ) from e

                if len(e.missing_properties) > 0:
                    raise ServiceValidationFailed(
                        f"Ledger connection validation failed; {e.missing_properties_error}"
                    ) from e

                if len(e.extra_properties) > 0:
                    self._warn(
                        "\n\t- ".join(
                            [
                                f"Following unknown ledgers found in the {component_id} override",
                                *e.extra_properties,
                            ]
                        )
                    )

    def validate_skill_config(self, skill_config: SkillConfig) -> None:
        """Check required overrides."""

        self.logger.info(f"Validating ABCI skill {skill_config.public_id}")
        model_params = skill_config.models.read("params")
        if model_params is None:
            raise ServiceValidationFailed(
                f"The chained ABCI skill `{skill_config.public_id}` does not contain `params` model"
            )

        self._validate_override(
            validator=ABCI_SKILL_MODEL_PARAMS_VALIDATOR,
            overrides=model_params.args,
            has_multiple_overrides=False,
            error_message="ABCI skill validation failed; {error}",
        )
        self.logger.info("No issues found in the ABCI skill configuration")

    def validate_agent_overrides(self, agent_config: AgentConfig) -> None:
        """Check required overrides."""

        self.logger.info(f"Validating agent {agent_config.public_id} overrides")
        errors = []
        for (
            component_id,
            component_config,
        ) in agent_config.component_configurations.items():
            try:
                self.validate_override(
                    component_id=component_id,
                    override=component_config,
                    has_multiple_overrides=False,
                )
            except ServiceValidationFailed as e:
                errors.append(str(e))

        if len(errors) > 0:
            error_string = "\n\t- ".join(errors)
            raise ServiceValidationFailed(
                "Agent overrides validation failed with following errors"
                f"\n\t- {error_string}"
            )

        self.logger.info("No issues found in the agent overrides")

    def validate_service_overrides(self) -> None:
        """Check required overrides."""

        self.logger.info("Validating service overrides")
        errors = []
        for _component_config in self.service_config.overrides:
            (
                component_config,
                component_id,
                has_multiple_overrides,
            ) = Service.process_metadata(configuration=_component_config.copy())
            try:
                self.validate_override(
                    component_id=component_id,
                    override=component_config,
                    has_multiple_overrides=has_multiple_overrides,
                )
            except ServiceValidationFailed as e:
                errors.append(str(e))

        if len(errors) > 0:
            error_string = "\n\t- ".join(errors)
            raise ServiceValidationFailed(
                "Service overrides validation failed with following errors"
                f"\n\t- {error_string}"
            )

        self.logger.info("No issues found in the service overrides")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Script to generate a markdown package table."""
from pathlib import Path
from typing import Dict

import yaml
from aea.cli.packages import get_package_manager


COL_WIDTH = 61

PACKAGE_TYPE_TO_CONFIG = {
    "agent": "aea-config.yaml",
    "connection": "connection.yaml",
    "contract": "contract.yaml",
    "service": "service.yaml",
    "skill": "skill.yaml",
    "protocol": "protocol.yaml",
}


def get_package_description(package: str) -> str:
    """Load the package description from its configuration file."""
    package_type, author, package_name, _ = package.split("/")
    config_path = Path(
        "packages",
        author,
        f"{package_type}s",
        package_name,
        PACKAGE_TYPE_TO_CONFIG[package_type],
    )
    with open(config_path, "r", encoding="utf-8") as config_file:
        print(config_path)
        config = yaml.load_all(config_file, Loader=yaml.FullLoader)
        return list(config)[0]["description"]


def get_packages() -> Dict[str, str]:
    """Get packages."""
    data = get_package_manager(Path("packages").relative_to(".")).json
    if "dev" in data:
        return {**data["dev"], **data["third_party"]}
    return data


def generate_table() -> None:
    """Generates a markdown table containing a package list"""

    data = get_packages()

    # Table header
    content = (
        f"| {'Package name'.ljust(COL_WIDTH, ' ')} | {'Package hash'.ljust(COL_WIDTH, ' ')} | {'Description'.ljust(COL_WIDTH * 2, ' ')} |\n"
        f"| {'-'*COL_WIDTH} | {'-'*COL_WIDTH} | {'-'*COL_WIDTH * 2} |\n"
    )

    # Table rows
    for package, package_hash in data.items():
        package_cell = package.ljust(COL_WIDTH, " ")
        hash_cell = f"`{package_hash}`".ljust(COL_WIDTH, " ")
        description = get_package_description(package)
        description_cell = f"{description}".ljust(COL_WIDTH * 2, " ")
        content += f"| {package_cell} | {hash_cell} | {description_cell} |\n"

    # Write table
    with open(
        Path("docs", "package_list.md"), mode="w", encoding="utf-8"
    ) as packages_list:
        packages_list.write(content)


if __name__ == "__main__":
    generate_table()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `autonomy analyse service` command"""

import logging
from copy import deepcopy
from typing import Any, Dict, List, Tuple
from unittest import mock

from aea.configurations.base import AgentConfig, SkillConfig
from aea.configurations.data_types import (
    ComponentId,
    ComponentType,
    PackageId,
    PackageType,
    PublicId,
)
from aea.helpers.cid import to_v0
from aea_cli_ipfs.ipfs_utils import IPFSDaemon, IPFSTool
from requests.exceptions import ConnectionError as RequestConnectionError

from autonomy.analyse.service import ServiceAnalyser
from autonomy.configurations.base import Service

from tests.test_autonomy.test_cli.base import BaseCliTest


DUMMY_SERVICE_HASH = "bafybeib56ojddzexxbapowofypmpk6zeznqaumwgj7ftneb5ua6sk5k5vj"
DUMMY_AGENT_HASH = "bafybeib56ojddzexxbapowofypmpk6zeznqaumwgj7ftneb5ua6sk5k5vm"
DUMMY_LEDGER_CONNECTION_HASH = (
    "bafybeib56ojddzexxbapowofypmpk6zeznqaumwgj7ftneb5ua6sk5k5vn"
)
DUMMY_ABCI_CONNECTION_HASH = (
    "bafybeib56ojddzexxbapowofypmpk6zeznqaumwgj7ftneb5ua6sk5k5vo"
)
DUMMY_SKILL_HASH = "bafybeib56ojddzexxbapowofypmpk6zeznqaumwgj7ftneb5ua6sk5k5vp"


def get_dummy_service_config() -> Dict:
    """Returns dummy service config."""
    return {
        "name": "dummy_service",
        "author": "valory",
        "version": "0.1.0",
        "description": "A dummy service config file with sigular overrides.",
        "aea_version": ">=1.0.0, <2.0.0",
        "license": "Apache-2.0",
        "fingerprint": {
            "README.md": "bafybeiapubcoersqnsnh3acia5hd7otzt7kjxekr6gkbrlumv6tkajl6jm"
        },
        "fingerprint_ignore_patterns": [],
        "agent": f"valory/dummy_agent:0.1.0:{DUMMY_AGENT_HASH}",
        "number_of_agents": 1,
    }


def get_dummy_overrides_skill(env_vars_with_name: bool = False) -> Dict:
    """Returns dummy skill overrides."""
    if env_vars_with_name:
        return {
            "public_id": "valory/abci_skill:0.1.0",
            "type": "skill",
            "models": {
                "params": {
                    "args": {
                        "message": "${MESSAGE:str:Hello, World!}",
                        "reset_pause_duration": "${RESET_PAUSE_DURATION:int:10}",
                        "setup": {
                            "safe_contract_address": "${SAFE_CONTRACT_ADDRESS:str:0xaddress}",
                            "all_participants": '${ALL_PARTICIPANTS:list:["0x"]}',
                            "consensus_threshold": "${CONSENSUS_THRESHOLD:int:null}",
                        },
                        "tendermint_url": "${TENDERMINT_URL:str:http://localhost}",
                        "tendermint_com_url": "${TENDERMINT_COM_URL:str:http://localhost}",
                        "tendermint_p2p_url": "${TENDERMINT_P2P_URL:str:http://localhost}",
                        "service_registry_address": "${SERVICE_REGISTRY_ADDRESS:str:0xaddress}",
                        "share_tm_config_on_startup": "${SHARE_TM_CONFIG_ON_STARTUP:bool:false}",
                        "on_chain_service_id": "${ON_CHAIN_SERVICE_ID:int:null}",
                    },
                }
            },
        }
    return {
        "public_id": "valory/abci_skill:0.1.0",
        "type": "skill",
        "models": {
            "params": {
                "args": {
                    "message": "${str:Hello, World!}",
                    "reset_pause_duration": "${int:10}",
                    "setup": {
                        "safe_contract_address": "${str:0xaddress}",
                        "all_participants": '${list:["0x"]}',
                        "consensus_threshold": "${int:null}",
                    },
                    "tendermint_url": "${str:http://localhost}",
                    "tendermint_com_url": "${str:http://localhost}",
                    "tendermint_p2p_url": "${str:http://localhost}",
                    "service_registry_address": "${str:0xaddress}",
                    "share_tm_config_on_startup": "${bool:false}",
                    "on_chain_service_id": "${int:null}",
                },
            }
        },
    }


def get_dummy_overrides_ledger_connection(env_vars_with_name: bool = False) -> Dict:
    """Returns dummy connection overrides"""
    if env_vars_with_name:
        return {
            "public_id": "valory/ledger:0.1.0",
            "type": "connection",
            "config": {
                "ledger_apis": {
                    "ethereum": {
                        "address": "${ADDRESS:str:address}",
                        "chain_id": "${CHAIN_ID:int:1}",
                        "poa_chain": "${POA_CHAIN:bool:false}",
                        "default_gas_price_strategy": "${DEFAULT_GAS_PRICE_STRATEGY:str:eip1559}",
                    }
                }
            },
        }
    return {
        "public_id": "valory/ledger:0.1.0",
        "type": "connection",
        "config": {
            "ledger_apis": {
                "ethereum": {
                    "address": "${str:address}",
                    "chain_id": "${int:1}",
                    "poa_chain": "${bool:false}",
                    "default_gas_price_strategy": "${str:eip1559}",
                }
            }
        },
    }


def get_dummy_overrides_abci_connection(env_vars_with_name: bool = False) -> Dict:
    """Returns dummy connection overrides"""
    if env_vars_with_name:
        return {
            "public_id": "valory/abci:0.1.0",
            "type": "connection",
            "config": {
                "host": "${ABCI_HOST:str:host}",
                "port": "${ABCI_PORT:str:port}",
                "use_tendermint": "${ABCI_USE_TENDERMINT:bool:false}",
            },
        }

    return {
        "public_id": "valory/abci:0.1.0",
        "type": "connection",
        "config": {
            "host": "${str:host}",
            "port": "${str:port}",
            "use_tendermint": "${bool:false}",
        },
    }


def get_dummy_agent_config() -> Dict:
    """Returns dummy agent config"""
    return {
        "agent_name": "dummy_agent",
        "author": "valory",
        "version": "0.1.0",
        "license": "Apache-2.0",
        "description": "agent",
        "aea_version": ">=1.0.0, <2.0.0",
        "fingerprint": {},
        "fingerprint_ignore_patterns": [],
        "connections": [
            f"valory/abci:0.1.0:{DUMMY_ABCI_CONNECTION_HASH}",
            f"valory/ledger:0.1.0:{DUMMY_LEDGER_CONNECTION_HASH}",
        ],
        "contracts": [],
        "protocols": [],
        "skills": [
            f"valory/abci_skill:0.1.0:{DUMMY_SKILL_HASH}",
        ],
        "default_ledger": "ethereum",
        "required_ledgers": ["ethereum"],
        "default_routing": {},
        "connection_private_key_paths": {},
        "private_key_paths": {},
        "logging_config": {
            "version": 1,
            "disable_existing_loggers": False,
        },
        "dependencies": {},
        "default_connection": None,
    }


def get_dummy_skill_config() -> Dict:
    """Returns dummy skill config"""
    return {
        "name": "abci_skill",
        "author": "valory",
        "version": "0.1.0",
        "type": "skill",
        "description": "Hello World ABCI application.",
        "license": "Apache-2.0",
        "aea_version": ">=1.0.0, <2.0.0",
        "fingerprint": {},
        "fingerprint_ignore_patterns": [],
        "connections": [],
        "contracts": [],
        "protocols": [],
        "skills": [
            "valory/abstract_round_abci:0.1.0:bafybeiami2y2uwvtmjf76hl3jvfahujdir5docjby6erdyjl66ekoblesi"
        ],
        "behaviours": {"main": {"args": {}, "class_name": "DummyBehaviour"}},
        "handlers": {},
        "models": {
            "params": {
                "args": {
                    "cleanup_history_depth": 1,
                    "cleanup_history_depth_current": None,
                    "drand_public_key": "drand_public_key",
                    "genesis_config": {},
                    "keeper_timeout": 30,
                    "max_attempts": 10,
                    "max_healthcheck": 120,
                    "observation_interval": 10,
                    "on_chain_service_id": None,
                    "request_retry_delay": 1,
                    "request_timeout": 10,
                    "reset_pause_duration": 10,
                    "reset_tendermint_after": 2,
                    "retry_attempts": 400,
                    "retry_timeout": 3,
                    "round_timeout_seconds": 30,
                    "service_id": "hello_world_abci",
                    "service_registry_address": None,
                    "setup": {
                        "all_participants": "- '0x0000000000000000000000000000000000000000'",
                        "safe_contract_address": "0x0000000000000000000000000000000000000000",
                        "consensus_threshold": None,
                    },
                    "share_tm_config_on_startup": False,
                    "sleep_time": 1,
                    "tendermint_check_sleep_delay": 3,
                    "tendermint_com_url": "http://localhost:8080",
                    "tendermint_max_retries": 5,
                    "tendermint_p2p_url": "localhost:26656",
                    "tendermint_url": "http://localhost:26657",
                    "tx_timeout": 10,
                    "use_termination": False,
                },
                "class_name": "Params",
            },
        },
        "dependencies": {},
        "is_abstract": False,
    }


class BaseAnalyseServiceTest(BaseCliTest):
    """Base class for service verifier tests"""

    cli_options: Tuple[str, ...] = ("analyse", "service")
    public_id_option = ("--public-id", "valory/service")
    token_id_option = ("--token-id", "1")

    @classmethod
    def setup_class(cls) -> None:
        """Setup class."""
        super().setup_class()

        cls.cli_runner.mix_stderr = False

    @staticmethod
    def patch_check_agent_dependencies_published(**kwargs: Any) -> mock._patch:
        """Patch `ServiceAnalyser.check_agent_dependencies_published` method"""
        return mock.patch.object(
            ServiceAnalyser, "check_agent_dependencies_published", **kwargs
        )

    @staticmethod
    def patch_verify_overrides(**kwargs: Any) -> mock._patch:
        """Patch `ServiceAnalyser.verify_overrides` method"""
        return mock.patch.object(ServiceAnalyser, "verify_overrides", **kwargs)

    @staticmethod
    def patch_validate_service_overrides(**kwargs: Any) -> mock._patch:
        """Patch `ServiceAnalyser.validate_service_overrides` method"""
        return mock.patch.object(
            ServiceAnalyser, "validate_service_overrides", **kwargs
        )

    @staticmethod
    def patch_check_skill_override(**kwargs: Any) -> mock._patch:
        """Patch `ServiceAnalyser.check_skill_override` method"""
        return mock.patch.object(ServiceAnalyser, "check_skill_override", **kwargs)

    @staticmethod
    def patch_ipfs_tool(pins: List[str]) -> mock._patch:
        """Patch `ServiceAnalyser.check_skill_override` method"""
        return mock.patch.object(IPFSTool, "all_pins", return_value=pins)

    @staticmethod
    def patch_check_on_chain_state(**kwargs: Any) -> mock._patch:
        """Patch `ServiceAnalyser.check_on_chain_state` method"""
        return mock.patch.object(ServiceAnalyser, "check_on_chain_state", **kwargs)

    @staticmethod
    def patch_ipfs_daemon_is_started(**kwargs: Any) -> mock._patch:
        """Patch IPFSDaemon.is_started method"""
        return mock.patch.object(
            IPFSDaemon,
            "is_started",
            **kwargs,
        )

    @staticmethod
    def patch_loader(
        service_data: List[Dict[str, Any]],
        agent_data: List[Dict[str, Any]],
        skill_data: Dict[str, Any],
        is_remote: bool = False,
    ) -> mock._patch:
        """Patch service loader method"""

        def _loader_patch(package_id: PackageId, **kargs: Any) -> Any:
            if package_id.package_type == PackageType.SERVICE:
                config, *overrides = service_data
                service = Service.from_json(config)
                service.overrides = overrides
                return service

            if package_id.package_type == PackageType.AGENT:
                config, *_overrides = agent_data
                agent_config = AgentConfig.from_json(config)
                if len(_overrides) > 0:
                    processed_overrides = {}
                    for o in _overrides:
                        processed_overrides[
                            ComponentId(
                                component_type=ComponentType(o.pop("type")),
                                public_id=PublicId.from_str(o.pop("public_id")),
                            )
                        ] = o
                    agent_config.component_configurations = processed_overrides
                return agent_config

            if package_id.package_type == PackageType.SKILL:
                return SkillConfig.from_json(skill_data)

        if is_remote:
            return mock.patch(
                "autonomy.cli.helpers.analyse._load_from_ipfs",
                new=_loader_patch,
            )

        return mock.patch(
            "autonomy.cli.helpers.analyse._load_from_local",
            new=_loader_patch,
        )

    @staticmethod
    def patch_get_on_chain_service_id() -> mock._patch:
        """Patch _get_on_chain_service_id"""

        return mock.patch(
            "autonomy.cli.helpers.analyse._get_on_chain_service_id",
            return_value=PackageId(
                package_type=PackageType.SERVICE,
                public_id=PublicId.from_str(f"valory/service:{DUMMY_SERVICE_HASH}"),
            ),
        )


class TestVerifySkillConfig(BaseAnalyseServiceTest):
    """Test verify overrides method."""

    def test_abci_skill_not_found(self) -> None:
        """Test run."""

        skill_config = get_dummy_skill_config()
        skill_config["skills"] = []
        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(env_vars_with_name=True),
            ],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=skill_config,
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert "Chained ABCI skill package not found" in result.stderr

    def test_params_model_not_found(self) -> None:
        """Test run."""

        skill_config = get_dummy_skill_config()
        del skill_config["models"]["params"]

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(env_vars_with_name=True),
            ],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=skill_config,
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "The chained ABCI skill `valory/abci_skill:0.1.0` does not contain `params` model"
            in result.stderr
        )


class TestCheckRequiredAgentOverrides(BaseAnalyseServiceTest):
    """Test agent override verification."""

    def test_missing_override(self) -> None:
        """Test missing override run."""

        skill_config_service = get_dummy_overrides_skill(env_vars_with_name=True)
        skill_config_agent = get_dummy_overrides_skill()

        del skill_config_agent["models"]["params"]["args"]["setup"]

        with self.patch_loader(
            service_data=[get_dummy_service_config(), skill_config_service],
            agent_data=[get_dummy_agent_config(), skill_config_agent],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "Error: Agent overrides validation failed with following errors"
            "\n\t- ABCI Skill `valory/abci_skill:0.1.0` override validation failed; Following properties are require but missing 'setup'\n"
            in result.stderr
        ), result.stdout


class TestCheckRequiredServiceOverrides(BaseAnalyseServiceTest):
    """Test service override verification."""

    def test_abci_skill_params_not_defined(self) -> None:
        """Test successful run."""

        skill_config = get_dummy_overrides_skill(env_vars_with_name=True)
        del skill_config["models"]["params"]

        with self.patch_loader(
            service_data=[get_dummy_service_config(), skill_config],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "Service overrides validation failed with following errors"
            "\n\t- ABCI Skill `valory/abci_skill:0.1.0` override validation failed; Following properties are require but missing 'params'\n"
            in result.stderr
        ), result.stdout

    def test_abci_skill_required_arg_not_defined(self, caplog: Any) -> None:
        """Test successful run."""

        skill_config = get_dummy_overrides_skill(env_vars_with_name=True)
        del skill_config["models"]["params"]["args"]["tendermint_url"]

        with self.patch_loader(
            service_data=[get_dummy_service_config(), skill_config],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]), caplog.at_level(logging.WARNING):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 0, result.stderr
        assert (
            "(agent, valory/dummy_agent:0.1.0) contains configuration which is missing from (service, valory/dummy_service:0.1.0)\n\tPath: models.params.args\n\tMissing parameters: {'tendermint_url'}"
            in caplog.text
        )
        assert "Path: models.params.args" in caplog.text
        assert "Missing parameters: {'tendermint_url'}" in caplog.text

    def test_abci_skill_setup_param_not_defined(self) -> None:
        """Test successful run."""

        skill_config = get_dummy_overrides_skill()

        del skill_config["models"]["params"]["args"]["setup"]["all_participants"]

        with self.patch_loader(
            service_data=[get_dummy_service_config(), skill_config],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "Service overrides validation failed with following errors\n\t- ABCI Skill `valory/abci_skill:0.1.0` override validation failed; Following properties are require but missing 'all_participants'\n"
            in result.stderr
        )

    def test_ledger_connection_ledger_not_defined(self) -> None:
        """Test successful run."""

        connection_config = get_dummy_overrides_ledger_connection(
            env_vars_with_name=True
        )
        del connection_config["config"]["ledger_apis"]["ethereum"]

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(env_vars_with_name=True),
                connection_config,
            ],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "Service overrides validation failed with following errors"
            "\n\t- (connection, valory/ledger:0.1.0) override needs at least one ledger API definition"
            in result.stderr
        )

    def test_ledger_connection_unknown_ledger_defined(self, caplog: Any) -> None:
        """Test successful run."""

        connection_config = get_dummy_overrides_ledger_connection(
            env_vars_with_name=True
        )
        connection_config["config"]["ledger_apis"]["solana"] = connection_config[
            "config"
        ]["ledger_apis"]["ethereum"]

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(env_vars_with_name=True),
                connection_config,
            ],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]), caplog.at_level(logging.WARNING):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "Following unknown ledgers found in the (connection, valory/ledger:0.1.0) override\n\t- solana\n"
            in caplog.text
        )

    def test_ledger_connection_param_not_defined(self) -> None:
        """Test successful run."""

        connection_config = get_dummy_overrides_ledger_connection()
        del connection_config["config"]["ledger_apis"]["ethereum"]["address"]

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(),
                connection_config,
            ],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "Service overrides validation failed with following errors"
            "\n\t- Ledger connection validation failed; Following properties are require but missing 'address'\n"
            in result.stderr
        )

    def test_abci_connection_setup_param_not_defined(self) -> None:
        """Test successful run."""

        connection_config = get_dummy_overrides_abci_connection()
        del connection_config["config"]["host"]

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(),
                connection_config,
            ],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "Service overrides validation failed with following errors"
            "\n\t- ABCI connection validation failed; Following properties are require but missing 'host'\n"
            in result.stderr
        )


class TestEnvVarValidation(BaseAnalyseServiceTest):
    """Test service override verification."""

    def test_list_var_validation(self) -> None:
        """Test agent var validation."""

        skill_config = get_dummy_overrides_skill()
        skill_config["models"]["params"]["args"]["messages"] = [
            {"text": "${str:hello}"}
        ]

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(env_vars_with_name=True),
            ],
            agent_data=[get_dummy_agent_config(), skill_config],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 0, result.stdout

    def test_agent_var_validation(self) -> None:
        """Test agent var validation."""

        skill_config = get_dummy_overrides_skill()
        skill_config["models"]["params"]["args"]["message"] = "Hello, World!"

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(env_vars_with_name=True),
            ],
            agent_data=[get_dummy_agent_config(), skill_config],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "(skill, valory/abci_skill:0.1.0) envrionment variable validation failed with following error"
            "\n\t- `models.params.args.message` needs to be defined as a environment variable"
            in result.stderr
        ), result.stdout

    def test_var_type_validation(self) -> None:
        """Test var type validation."""

        skill_config = get_dummy_overrides_skill()
        skill_config["models"]["params"]["args"]["tendermint_url"] = None

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(env_vars_with_name=True),
            ],
            agent_data=[get_dummy_agent_config(), skill_config],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "(skill, valory/abci_skill:0.1.0) envrionment variable validation failed with following error"
            "\n\t- `models.params.args.tendermint_url` needs to be defined as a environment variable"
            in result.stderr
        ), result.stdout

    def test_loading(self) -> None:
        """Test var type validation."""

        skill_config = get_dummy_overrides_skill()
        skill_config["models"]["params"]["args"]["tendermint_url"] = "${int:hello}"

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(env_vars_with_name=True),
            ],
            agent_data=[get_dummy_agent_config(), skill_config],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "(skill, valory/abci_skill:0.1.0) envrionment variable validation failed with following error"
            "\n\t- `models.params.args.tendermint_url` validation failed with following error; Cannot convert string `hello` to type `int`"
            in result.stderr
        ), result.stdout


class TestCheckAgentDependenciesPublished(BaseAnalyseServiceTest):
    """Test `check_agent_package_published` method."""

    def test_dependency_not_found(self) -> None:
        """Test failure."""

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(env_vars_with_name=True),
            ],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
            is_remote=True,
        ), self.patch_ipfs_tool(
            []
        ), self.patch_validate_service_overrides(), self.patch_ipfs_daemon_is_started(
            return_value=True,
        ), self.patch_get_on_chain_service_id(), self.patch_check_on_chain_state():
            result = self.run_cli(commands=self.token_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "Error: Package required for service valory/dummy_service:0.1.0 is not published on the IPFS registry"
            in result.stderr
        )
        assert "agent: valory/dummy_agent:0.1.0" in result.stderr


class TestVerifyCrossOverrides(BaseAnalyseServiceTest):
    """Test verify overrides method."""

    def test_overrides_missing_from_agent(self) -> None:
        """Test missing override section from agent config."""

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                get_dummy_overrides_skill(env_vars_with_name=True),
                get_dummy_overrides_ledger_connection(env_vars_with_name=True),
            ],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool([]), self.patch_validate_service_overrides():
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "Overrides with following component IDs are defined in the service configuration but not in the agent configuration"
            "\n\t- (connection, valory/ledger:0.1.0)\n" in result.stderr
        )

    def test_overrides_missing_from_service(self, caplog: Any) -> None:
        """Test missing override section from service config."""

        with self.patch_loader(
            service_data=[get_dummy_service_config()],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
        ), self.patch_ipfs_tool(
            []
        ), self.patch_validate_service_overrides(), caplog.at_level(
            logging.WARNING
        ):
            result = self.run_cli(commands=self.public_id_option)

        assert result.exit_code == 0, result.stdout
        assert (
            "Overrides with following component IDs are defined in the agent configuration but not in the service configuration"
            "\n\t- (skill, valory/abci_skill:0.1.0)" in caplog.text
        )


class TestCheckOnChainState(BaseAnalyseServiceTest):
    """Test `check_on_chain_state` method."""

    def test_on_chain_status_check(self, caplog: Any) -> None:
        """Test `check_on_chain_state` failure"""

        with self.patch_loader(
            service_data=[get_dummy_service_config(), get_dummy_overrides_skill()],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
            is_remote=True,
        ), mock.patch(
            "autonomy.analyse.service.get_service_info", return_value=[None, 0, None]
        ), self.patch_ipfs_tool(
            []
        ), self.patch_ipfs_daemon_is_started(
            return_value=True,
        ), self.patch_get_on_chain_service_id(), caplog.at_level(
            logging.WARNING
        ):
            result = self.run_cli(commands=self.token_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "Service needs to be in deployed state on-chain; Current state=ServiceState.NON_EXISTENT"
            in caplog.text
        )

    def test_on_chain_status_check_rpc_failure(self) -> None:
        """Test `check_on_chain_state` RPC connection failure"""

        with self.patch_loader(
            service_data=[get_dummy_service_config(), get_dummy_overrides_skill()],
            agent_data=[get_dummy_agent_config(), get_dummy_overrides_skill()],
            skill_data=get_dummy_skill_config(),
            is_remote=True,
        ), mock.patch(
            "autonomy.analyse.service.get_service_info",
            side_effect=RequestConnectionError,
        ), self.patch_ipfs_tool(
            []
        ), self.patch_ipfs_daemon_is_started(
            return_value=True,
        ), self.patch_get_on_chain_service_id():
            result = self.run_cli(commands=self.token_id_option)

        assert result.exit_code == 1, result.stdout
        assert (
            "On chain service validation failed, could not connect to the RPC"
            in result.stderr
        )


class TestCheckSuccessful(BaseAnalyseServiceTest):
    """Test a successful check"""

    def test_run(self) -> None:
        """Test run."""

        skill_config = get_dummy_overrides_skill(env_vars_with_name=True)
        models = skill_config.pop("models")
        skill_config_original = get_dummy_skill_config()
        skill_config_original["models"]["benchmark_tool"] = {
            "args": {"log_dir": "/logs"}
        }

        for i in range(4):
            skill_config[i] = {"models": deepcopy(models)}

        with self.patch_loader(
            service_data=[
                get_dummy_service_config(),
                skill_config,
                get_dummy_overrides_abci_connection(env_vars_with_name=True),
                get_dummy_overrides_ledger_connection(env_vars_with_name=True),
            ],
            agent_data=[
                get_dummy_agent_config(),
                get_dummy_overrides_skill(),
                get_dummy_overrides_abci_connection(),
                get_dummy_overrides_ledger_connection(),
            ],
            skill_data=skill_config_original,
            is_remote=True,
        ), self.patch_ipfs_tool(
            pins=list(
                map(
                    to_v0,
                    [
                        DUMMY_AGENT_HASH,
                        DUMMY_LEDGER_CONNECTION_HASH,
                        DUMMY_ABCI_CONNECTION_HASH,
                        DUMMY_SKILL_HASH,
                    ],
                )
            )
        ), self.patch_ipfs_daemon_is_started(
            return_value=True,
        ), self.patch_get_on_chain_service_id(), mock.patch(
            "autonomy.analyse.service.get_service_info",
            return_value=(None, 4, None),
        ):
            result = self.run_cli(commands=self.token_id_option)

        assert result.exit_code == 0, result.stderr
        assert "Service is ready to be deployed" in result.output



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test utils."""

import re
from typing import Dict
from unittest import mock

import pytest
from aea.configurations.data_types import PackageId, PublicId

from autonomy.chain.exceptions import DependencyError
from autonomy.chain.utils import (
    get_ipfs_hash_from_uri,
    parse_public_id_from_metadata,
    verify_component_dependencies,
    verify_service_dependencies,
)

from tests.test_autonomy.test_chain.base import DUMMY_HASH


def get_dummy_metadata(
    name: str = "valory/package",
    description: str = "description",
    code_uri: str = "code_uri",
    nft_hash: str = "nft_hash",
    version: str = "0.1.0",
) -> Dict:
    """Get dummy metadata."""

    return {
        "name": name,
        "description": description,
        "code_uri": code_uri,
        "image": nft_hash,
        "attributes": [{"trait_type": "version", "value": version}],
    }


def test_get_ipfs_hash_from_uri() -> None:
    """Test `get_ipfs_hash_from_uri` method"""

    assert get_ipfs_hash_from_uri("ipfs://SOME_HASH") == "SOME_HASH"


def test_verify_component_dependencies_no_dep_found_locally() -> None:
    """Test `verify_component_dependencies`"""

    with mock.patch(
        "autonomy.chain.utils.resolve_component_id", return_value=get_dummy_metadata()
    ), pytest.raises(
        DependencyError,
        match="On chain dependency with id 0 and public ID valory/package:any not found in the local package configuration",
    ):
        verify_component_dependencies(
            ledger_api=mock.MagicMock(),
            contract_address="0xdummy_contract",
            dependencies=[
                0,
            ],
            package_configuration=mock.MagicMock(package_dependencies=[]),
        )


def test_verify_component_dependencies_no_dep_found_on_chain() -> None:
    """Test `verify_component_dependencies`"""

    with pytest.raises(
        DependencyError,
        match=re.escape(
            "Please provide on chain ID as dependency for following packages; ['author/package:any']"
        ),
    ):
        verify_component_dependencies(
            ledger_api=mock.MagicMock(),
            contract_address="0xdummy_contract",
            dependencies=[],
            package_configuration=mock.MagicMock(
                package_dependencies=[
                    PackageId.from_uri_path("skill/author/package/0.1.0").with_hash(
                        DUMMY_HASH
                    )
                ]
            ),
        )


def test_verify_component_dependencies_hash_dont_match() -> None:
    """Test `verify_component_dependencies`"""

    with mock.patch(
        "autonomy.chain.utils.resolve_component_id",
        return_value=get_dummy_metadata(),
    ), pytest.raises(
        DependencyError,
        match="Package hash does not match for the on chain package and the local package",
    ):
        verify_component_dependencies(
            ledger_api=mock.MagicMock(),
            contract_address="0xdummy_contract",
            dependencies=[
                0,
            ],
            package_configuration=mock.MagicMock(
                package_dependencies=[
                    PackageId.from_uri_path("skill/valory/package/0.1.0").with_hash(
                        DUMMY_HASH
                    )
                ]
            ),
        )


def test_verify_component_dependencies_multiple_component_with_same_public_id() -> None:
    """Test `verify_component_dependencies`"""

    phash_1 = DUMMY_HASH.replace("0", "1")
    phash_2 = DUMMY_HASH.replace("0", "2")

    with mock.patch(
        "autonomy.chain.utils.resolve_component_id",
        side_effect=[
            get_dummy_metadata(code_uri=phash_1),
            get_dummy_metadata(code_uri=phash_2),
        ],
    ):
        verify_component_dependencies(
            ledger_api=mock.MagicMock(),
            contract_address="0xdummy_contract",
            dependencies=[0, 1],
            package_configuration=mock.MagicMock(
                package_dependencies=[
                    PackageId.from_uri_path(
                        "connection/valory/package/0.1.0",
                    ).with_hash(phash_1),
                    PackageId.from_uri_path(
                        "contract/valory/package/0.1.0",
                    ).with_hash(phash_2),
                ]
            ),
        )


def test_verify_component_dependencies_multiple_component_with_same_public_id_skip_hash_check() -> (
    None
):
    """Test `verify_component_dependencies`"""

    phash_1 = DUMMY_HASH.replace("0", "1")
    phash_2 = DUMMY_HASH.replace("0", "2")

    with mock.patch(
        "autonomy.chain.utils.resolve_component_id",
        side_effect=[
            get_dummy_metadata(code_uri=phash_1),
            get_dummy_metadata(code_uri=phash_2),
        ],
    ):
        verify_component_dependencies(
            ledger_api=mock.MagicMock(),
            contract_address="0xdummy_contract",
            dependencies=[0, 1],
            package_configuration=mock.MagicMock(
                package_dependencies=[
                    PackageId.from_uri_path(
                        "connection/valory/package/0.1.0",
                    ).with_hash(phash_1),
                    PackageId.from_uri_path(
                        "contract/valory/package/0.1.0",
                    ).with_hash(phash_2),
                ]
            ),
            skip_hash_check=True,
        )


def test_verify_component_dependencies_multiple_component_with_same_public_id_fail() -> (
    None
):
    """Test `verify_component_dependencies`"""

    phash_1 = DUMMY_HASH.replace("0", "1")
    phash_2 = DUMMY_HASH.replace("0", "2")

    with mock.patch(
        "autonomy.chain.utils.resolve_component_id",
        side_effect=[
            get_dummy_metadata(code_uri=phash_1),
            get_dummy_metadata(code_uri="phash_2"),
        ],
    ), pytest.raises(
        DependencyError,
        match="Package hash does not match for the on chain package and the local package; Dependency=1",
    ):
        verify_component_dependencies(
            ledger_api=mock.MagicMock(),
            contract_address="0xdummy_contract",
            dependencies=[0, 1],
            package_configuration=mock.MagicMock(
                package_dependencies=[
                    PackageId.from_uri_path(
                        "connection/valory/package/0.1.0",
                    ).with_hash(phash_1),
                    PackageId.from_uri_path(
                        "contract/valory/package/0.1.0",
                    ).with_hash(phash_2),
                ]
            ),
        )


def test_verify_service_dependencies_dep_not_found() -> None:
    """Test verify `verify_service_dependencies` method"""

    with mock.patch(
        "autonomy.chain.utils.resolve_component_id", return_value=get_dummy_metadata()
    ), pytest.raises(
        DependencyError,
        match="On chain ID of the agent does not match with the one in the service configuration",
    ):
        verify_service_dependencies(
            ledger_api=mock.MagicMock(),
            contract_address="0xdummy_contract",
            agent_id=0,
            service_configuration=mock.MagicMock(),
        )


def test_verify_service_dependencies_hash_dont_match() -> None:
    """Test verify `verify_service_dependencies` method"""

    with mock.patch(
        "autonomy.chain.utils.resolve_component_id", return_value=get_dummy_metadata()
    ), pytest.raises(
        DependencyError,
        match="Package hash does not match for the on chain package and the local package",
    ):
        verify_service_dependencies(
            ledger_api=mock.MagicMock(),
            contract_address="0xdummy_contract",
            agent_id=0,
            service_configuration=mock.MagicMock(
                agent=PublicId("valory", "package", package_hash=DUMMY_HASH)
            ),
        )


@pytest.mark.parametrize(
    "public_id_string",
    [
        "author/package_name",
        "component_type/author/name",
        "skill/author/name/0.1.0",
    ],
)
def test_parse_public_id_from_metadata(public_id_string: str) -> None:
    """Test verify `parse_public_id_from_metadata` method"""

    public_id = parse_public_id_from_metadata(
        id_string=public_id_string,
    )

    assert isinstance(public_id, PublicId)


def test_parse_public_id_from_metadata_fail() -> None:
    """Test verify `parse_public_id_from_metadata` method"""

    with pytest.raises(DependencyError, match="Invalid package name found `public_id`"):
        parse_public_id_from_metadata(
            id_string="public_id",
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the payloads.py module of the skill."""

# pylint: skip-file

from packages.valory.skills.reset_pause_abci.payloads import ResetPausePayload


def test_reset_pause_payload() -> None:
    """Test `ResetPausePayload`."""

    payload = ResetPausePayload(sender="sender", period_count=1)

    assert payload.period_count == 1
    assert payload.data == {"period_count": 1}
    assert ResetPausePayload.from_json(payload.json) == payload



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test fetch command."""

import os
import shutil
from pathlib import Path
from unittest import mock

from aea.cli.fetch import NotAnAgentPackage
from aea.configurations.constants import (
    DEFAULT_README_FILE,
    DEFAULT_SERVICE_CONFIG_FILE,
)
from aea.configurations.loader import ConfigLoader
from aea.helpers.base import cd
from aea.helpers.io import open_file

from autonomy.cli.helpers.registry import IPFSTool
from autonomy.configurations.base import Service

from tests.conftest import ROOT_DIR
from tests.test_autonomy.base import get_dummy_service_config
from tests.test_autonomy.test_cli.base import BaseCliTest, cli


IPFS_REGISTRY = "/dns/registry.autonolas.tech/tcp/443/https"


class FetchTest(BaseCliTest):
    """FetchTest base class"""

    packages_dir: Path
    package_type: str  # agent or service

    def setup(self) -> None:
        """Setup class."""

        super().setup()

        self.packages_dir = self.t / "packages"
        self.cli_options = (
            "--registry-path",
            str(self.packages_dir),
            "fetch",
            f"--{self.package_type}",
        )

        shutil.copytree(ROOT_DIR / "packages", self.packages_dir)
        os.chdir(self.t)


class TestFetchAgentCommand(FetchTest):
    """Test fetch agent command"""

    package_type = "agent"

    def test_fetch_agent(self) -> None:
        """Test fetch agent"""

        result = self.run_cli(("--local", "valory/counter"))
        assert result.exit_code == 0
        assert "Agent counter successfully fetched." in result.stdout

    def test_not_an_agent_package_raises(self) -> None:
        """Test fetch agent"""

        expected = "Error: Downloaded packages is not an agent package, if you intend to download a service please use `--service` flag or check the hash"
        with mock.patch("autonomy.cli.fetch.do_fetch", side_effect=NotAnAgentPackage):
            result = self.run_cli(("--remote", "valory/counter"))
            assert result.exit_code == 1
            assert expected in result.stdout


class TestFetchServiceCommand(FetchTest):
    """Test fetch service command."""

    package_type = "service"

    def test_fetch_service_local(
        self,
    ) -> None:
        """Test fetch service."""

        service = self.t / "counter"
        result = self.run_cli(("--local", "valory/counter"))

        assert result.exit_code == 0, result.output
        assert service.exists()

        result = self.run_cli(("--local", "valory/counter"))
        assert result.exit_code == 1, result.output
        assert (
            'Item "counter" already exists in target folder' in result.output
        ), result.output

        shutil.rmtree(service)

    def test_publish_and_fetch_service_ipfs(self) -> None:
        """Test fetch service."""
        expected_hash = "bafybeicojucnpgeud7lmomfkppwbnxpnbtfwvqritalkzessfjbcze27re"

        service_dir = self.t / "dummy_service"
        service_file = service_dir / DEFAULT_SERVICE_CONFIG_FILE
        service_dir.mkdir()
        with open_file(service_file, "w+") as fp:
            service_conf, *overrides = get_dummy_service_config()
            service_conf["overrides"] = overrides
            service = Service.from_json(service_conf)
            ConfigLoader(Service.schema, Service).dump(service, fp)

        (service_dir / DEFAULT_README_FILE).write_text("Dummy Service")

        with mock.patch(
            "autonomy.cli.helpers.registry.get_default_remote_registry",
            new=lambda: "ipfs",
        ), mock.patch(
            "autonomy.cli.helpers.registry.get_ipfs_node_multiaddr",
            new=lambda: IPFS_REGISTRY,
        ), cd(
            service_dir
        ):
            result = self.cli_runner.invoke(cli, ["publish", "--remote"])

            assert result.exit_code == 0, result.output
            assert expected_hash in result.output

        with mock.patch(
            "autonomy.cli.helpers.registry.get_default_remote_registry",
            new=lambda: "http",
        ), cd(service_dir):
            result = self.run_cli(("--remote", expected_hash))
            assert result.exit_code == 1, result.output
            assert "HTTP registry not supported." in result.output, result.output

        with mock.patch(
            "autonomy.cli.helpers.registry.get_default_remote_registry",
            new=lambda: "ipfs",
        ), mock.patch(
            "autonomy.cli.helpers.registry.get_ipfs_node_multiaddr",
            new=lambda: IPFS_REGISTRY,
        ), cd(
            service_dir
        ):
            result = self.run_cli(("--remote", expected_hash))
            assert result.exit_code == 0, result.output
            assert service_dir.exists()

        alias = "some_service"

        with mock.patch(
            "autonomy.cli.helpers.registry.get_default_remote_registry",
            new=lambda: "ipfs",
        ), mock.patch(
            "autonomy.cli.helpers.registry.get_ipfs_node_multiaddr",
            new=lambda: IPFS_REGISTRY,
        ), cd(
            service_dir.parent
        ):
            result = self.run_cli(("--remote", "--alias", alias, expected_hash))
            assert result.exit_code == 0, result.output
            assert (service_dir.parent / alias).exists()

        shutil.rmtree(service_dir)

    def test_fetch_service_mixed(
        self,
    ) -> None:
        """Test fetch service in mixed mode."""
        with mock.patch(
            "autonomy.cli.helpers.registry.fetch_service_local",
            side_effect=Exception("expected"),
        ) as fetch_local_mock, mock.patch(
            "autonomy.cli.helpers.registry.fetch_service_ipfs"
        ) as fetch_remote_mock:
            result = self.run_cli(("--mixed", "valory/counter"))

        assert result.exit_code == 0, result.output
        fetch_local_mock.assert_called_once()
        fetch_remote_mock.assert_called_once()

    def test_not_a_service_package(
        self,
    ) -> None:
        """Test fetch service."""
        with mock.patch(
            "autonomy.cli.helpers.registry.get_default_remote_registry",
            new=lambda: "ipfs",
        ), mock.patch(
            "autonomy.cli.helpers.registry.get_ipfs_node_multiaddr",
            new=lambda: IPFS_REGISTRY,
        ), mock.patch.object(
            IPFSTool, "download", return_value=self.t
        ):
            result = self.run_cli(
                (
                    "--remote",
                    "bafybeicqvwvogloyw2ujhedbwv4opn2ngus6dh7ocxg7umhhawcnzpibrq",
                )
            )
            assert result.exit_code == 1, result.output
            assert (
                "Downloaded packages is not a service package, "
                "if you intend to download an agent please use "
                "`--agent` flag or check the hash"
            ) in result.output



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the behaviours.py module of the skill."""

# pylint: skip-file

from abc import ABC
from pathlib import Path
from typing import Any, Dict, Generator, Optional, Tuple
from unittest import mock
from unittest.mock import MagicMock

import pytest

from packages.valory.skills.abstract_round_abci import PUBLIC_ID
from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AbciApp,
    AbstractRound,
    BaseSynchronizedData,
    BaseTxPayload,
    DegenerateRound,
    EventType,
    RoundSequence,
)
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    BaseBehaviour,
    DegenerateBehaviour,
    TmManager,
)
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    _MetaRoundBehaviour,
)
from packages.valory.skills.abstract_round_abci.models import TendermintRecoveryParams


BEHAVIOUR_A_ID = "behaviour_a"
BEHAVIOUR_B_ID = "behaviour_b"
BEHAVIOUR_C_ID = "behaviour_c"
CONCRETE_BACKGROUND_BEHAVIOUR_ID = "background_behaviour"
ROUND_A_ID = "round_a"
ROUND_B_ID = "round_b"
CONCRETE_BACKGROUND_ROUND_ID = "background_round"


def test_skill_public_id() -> None:
    """Test skill module public ID"""

    assert PUBLIC_ID.name == Path(__file__).parents[1].name
    assert PUBLIC_ID.author == Path(__file__).parents[3].name


class RoundA(AbstractRound):
    """Round A."""

    round_id = ROUND_A_ID
    payload_class = BaseTxPayload
    payload_attribute = ""
    synchronized_data_class = BaseSynchronizedData

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, EventType]]:
        """End block."""

    def check_payload(self, payload: BaseTxPayload) -> None:
        """Check payload."""

    def process_payload(self, payload: BaseTxPayload) -> None:
        """Process payload."""


class RoundB(AbstractRound):
    """Round B."""

    round_id = ROUND_B_ID
    payload_class = BaseTxPayload
    payload_attribute = ""
    synchronized_data_class = BaseSynchronizedData

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, EventType]]:
        """End block."""

    def check_payload(self, payload: BaseTxPayload) -> None:
        """Check payload."""

    def process_payload(self, payload: BaseTxPayload) -> None:
        """Process payload."""


class ConcreteBackgroundRound(AbstractRound):
    """Concrete Background Round."""

    round_id = ROUND_B_ID
    payload_class = BaseTxPayload
    payload_attribute = ""
    synchronized_data_class = BaseSynchronizedData

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, EventType]]:
        """End block."""

    def check_payload(self, payload: BaseTxPayload) -> None:
        """Check payload."""

    def process_payload(self, payload: BaseTxPayload) -> None:
        """Process payload."""


class BehaviourA(BaseBehaviour):
    """Dummy behaviour."""

    behaviour_id = BEHAVIOUR_A_ID
    matching_round = RoundA

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize behaviour."""
        super().__init__(*args, **kwargs)
        self.count = 0

    def setup(self) -> None:
        """Setup behaviour."""
        self.count += 1

    def async_act(self) -> Generator:
        """Dummy act method."""
        yield


class BehaviourB(BaseBehaviour):
    """Dummy behaviour."""

    behaviour_id = BEHAVIOUR_B_ID
    matching_round = RoundB

    def async_act(self) -> Generator:
        """Dummy act method."""
        yield


class BehaviourC(BaseBehaviour, ABC):
    """Dummy behaviour."""

    matching_round = MagicMock()


def test_auto_behaviour_id() -> None:
    """Test that the 'auto_behaviour_id()' method works as expected."""

    assert BehaviourB.auto_behaviour_id() == BEHAVIOUR_B_ID
    assert BehaviourB.behaviour_id == BEHAVIOUR_B_ID
    assert BehaviourC.auto_behaviour_id() == "behaviour_c"
    assert isinstance(BehaviourC.behaviour_id, property)


class ConcreteBackgroundBehaviour(BaseBehaviour):
    """Dummy behaviour."""

    behaviour_id = CONCRETE_BACKGROUND_BEHAVIOUR_ID
    matching_round = ConcreteBackgroundRound

    def async_act(self) -> Generator:
        """Dummy act method."""
        yield


class ConcreteAbciApp(AbciApp):
    """Concrete ABCI App."""

    initial_round_cls = RoundA
    transition_function = {RoundA: {MagicMock(): RoundB}}
    event_to_timeout: Dict = {}


class ConcreteRoundBehaviour(AbstractRoundBehaviour):
    """Concrete round behaviour."""

    abci_app_cls = ConcreteAbciApp
    behaviours = {BehaviourA, BehaviourB}  # type: ignore
    initial_behaviour_cls = BehaviourA
    background_behaviour_cls = ConcreteBackgroundBehaviour


class TestAbstractRoundBehaviour:
    """Test 'AbstractRoundBehaviour' class."""

    def setup(self) -> None:
        """Set up the tests."""
        self.round_sequence_mock = MagicMock()
        context_mock = MagicMock(params=MagicMock())
        context_mock.state.round_sequence = self.round_sequence_mock
        context_mock.state.round_sequence.syncing_up = False
        self.round_sequence_mock.block_stall_deadline_expired = False
        self.behaviour = ConcreteRoundBehaviour(name="", skill_context=context_mock)
        self.behaviour.tm_manager = self.behaviour.instantiate_behaviour_cls(TmManager)  # type: ignore

    @pytest.mark.parametrize("use_termination", (True, False))
    def test_setup(self, use_termination: bool) -> None:
        """Test 'setup' method."""
        assert self.behaviour.background_behaviour is None
        self.behaviour.context.params.use_termination = use_termination
        self.behaviour.setup()
        background_behaviour = self.behaviour.background_behaviour
        assert self.behaviour.background_behaviour_cls == ConcreteBackgroundBehaviour
        assert (
            isinstance(background_behaviour, self.behaviour.background_behaviour_cls)
            if use_termination
            else background_behaviour is None
        )

    def test_teardown(self) -> None:
        """Test 'teardown' method."""
        self.behaviour.teardown()

    def test_current_behaviour_return_none(self) -> None:
        """Test 'current_behaviour' property return None."""
        assert self.behaviour.current_behaviour is None

    def test_act_current_behaviour_name_is_none(self) -> None:
        """Test 'act' with current behaviour None."""
        self.behaviour.current_behaviour = None
        with mock.patch.object(self.behaviour, "_process_current_round"):
            self.behaviour.act()

    def test_check_matching_round_consistency(self) -> None:
        """Test classmethod '_get_behaviour_id_to_behaviour_mapping', negative case."""
        rounds = [
            MagicMock(**{"auto_round_id.return_value": f"round_{i}"}) for i in range(3)
        ]
        mock_behaviours = [
            MagicMock(matching_round=round, behaviour_id=f"behaviour_{i}")
            for i, round in enumerate(rounds)
        ]

        with mock.patch.object(
            _MetaRoundBehaviour, "_check_all_required_classattributes_are_set"
        ), mock.patch.object(
            _MetaRoundBehaviour, "_check_behaviour_id_uniqueness"
        ), mock.patch.object(
            _MetaRoundBehaviour, "_check_initial_behaviour_in_set_of_behaviours"
        ), pytest.raises(
            ABCIAppInternalError,
            match="internal error: round round_0 is a final round it shouldn't have any matching behaviours",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = MagicMock(
                    get_all_round_classes=lambda include_termination_rounds: rounds,
                    final_states={
                        rounds[0],
                    },
                )
                behaviours = mock_behaviours  # type: ignore
                initial_behaviour_cls = MagicMock()

    def test_get_behaviour_id_to_behaviour_mapping_negative(self) -> None:
        """Test classmethod '_get_behaviour_id_to_behaviour_mapping', negative case."""
        behaviour_id = "behaviour_id"
        behaviour_1 = MagicMock(**{"auto_behaviour_id.return_value": behaviour_id})
        behaviour_2 = MagicMock(**{"auto_behaviour_id.return_value": behaviour_id})

        with pytest.raises(
            ValueError,
            match=f"cannot have two behaviours with the same id; got {behaviour_2} and {behaviour_1} both with id '{behaviour_id}'",
        ):
            with mock.patch.object(_MetaRoundBehaviour, "_check_consistency"):

                class MyRoundBehaviour(AbstractRoundBehaviour):
                    abci_app_cls = MagicMock
                    behaviours = [behaviour_1, behaviour_2]  # type: ignore
                    initial_behaviour_cls = MagicMock()

                MyRoundBehaviour(name=MagicMock(), skill_context=MagicMock())

    def test_get_round_to_behaviour_mapping_two_behaviours_same_round(self) -> None:
        """Test classmethod '_get_round_to_behaviour_mapping' when two different behaviours point to the same round."""
        behaviour_id_1 = "behaviour_id_1"
        behaviour_id_2 = "behaviour_id_2"
        round_cls = RoundA
        round_id = round_cls.auto_round_id()
        behaviour_1 = MagicMock(
            matching_round=round_cls,
            **{"auto_behaviour_id.return_value": behaviour_id_1},
        )
        behaviour_2 = MagicMock(
            matching_round=round_cls,
            **{"auto_behaviour_id.return_value": behaviour_id_2},
        )

        with pytest.raises(
            ValueError,
            match=f"the behaviours '{behaviour_2.auto_behaviour_id()}' and '{behaviour_1.auto_behaviour_id()}' point to the same matching round '{round_id}'",
        ):
            with mock.patch.object(_MetaRoundBehaviour, "_check_consistency"):

                class MyRoundBehaviour(AbstractRoundBehaviour):
                    abci_app_cls = ConcreteAbciApp
                    behaviours = [behaviour_1, behaviour_2]  # type: ignore
                    initial_behaviour_cls = behaviour_1

                MyRoundBehaviour(name=MagicMock(), skill_context=MagicMock())

    def test_get_round_to_behaviour_mapping_with_final_rounds(self) -> None:
        """Test classmethod '_get_round_to_behaviour_mapping' with final rounds."""

        class FinalRound(DegenerateRound, ABC):
            """A final round for testing."""

        behaviour_id_1 = "behaviour_id_1"
        behaviour_1 = MagicMock(behaviour_id=behaviour_id_1, matching_round=RoundA)

        class AbciAppTest(AbciApp):
            """Abci App for testing."""

            initial_round_cls = RoundA
            transition_function = {RoundA: {MagicMock(): FinalRound}, FinalRound: {}}
            event_to_timeout: Dict = {}
            final_states = {FinalRound}

        class MyRoundBehaviour(AbstractRoundBehaviour):
            abci_app_cls = AbciAppTest
            behaviours = {behaviour_1}
            initial_behaviour_cls = behaviour_1
            matching_round = FinalRound

        behaviour = MyRoundBehaviour(name=MagicMock(), skill_context=MagicMock())
        final_behaviour = behaviour._round_to_behaviour[FinalRound]
        assert issubclass(final_behaviour, DegenerateBehaviour)
        assert (
            final_behaviour.auto_behaviour_id()
            == f"degenerate_behaviour_{FinalRound.auto_round_id()}"
        )

    def test_check_behaviour_id_uniqueness_negative(self) -> None:
        """Test metaclass method '_check_consistency', negative case."""
        behaviour_id = "behaviour_id"
        behaviour_1_cls_name = "Behaviour1"
        behaviour_2_cls_name = "Behaviour2"
        behaviour_1 = MagicMock(
            __name__=behaviour_1_cls_name,
            **{"auto_behaviour_id.return_value": behaviour_id},
        )
        behaviour_2 = MagicMock(
            __name__=behaviour_2_cls_name,
            **{"auto_behaviour_id.return_value": behaviour_id},
        )

        with pytest.raises(
            ABCIAppInternalError,
            match=rf"behaviours \['{behaviour_1_cls_name}', '{behaviour_2_cls_name}'\] have the same behaviour id '{behaviour_id}'",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = MagicMock
                behaviours = [behaviour_1, behaviour_2]  # type: ignore
                initial_behaviour_cls = MagicMock()

    def test_check_consistency_two_behaviours_same_round(self) -> None:
        """Test metaclass method '_check_consistency' when two different behaviours point to the same round."""
        behaviour_id_1 = "behaviour_id_1"
        behaviour_id_2 = "behaviour_id_2"
        round_cls = RoundA
        round_id = round_cls.auto_round_id()
        behaviour_1 = MagicMock(
            matching_round=round_cls,
            **{"auto_behaviour_id.return_value": "behaviour_id_1"},
        )
        behaviour_2 = MagicMock(
            matching_round=round_cls,
            **{"auto_behaviour_id.return_value": "behaviour_id_2"},
        )

        with pytest.raises(
            ABCIAppInternalError,
            match=rf"internal error: behaviours \['{behaviour_id_1}', '{behaviour_id_2}'\] have the same matching round '{round_id}'",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = ConcreteAbciApp
                behaviours = [behaviour_1, behaviour_2]  # type: ignore
                initial_behaviour_cls = behaviour_1

    def test_check_initial_behaviour_in_set_of_behaviours_negative_case(self) -> None:
        """Test classmethod '_check_initial_behaviour_in_set_of_behaviours' when initial behaviour is NOT in the set."""
        behaviour_1 = MagicMock(
            matching_round=MagicMock(),
            **{"auto_behaviour_id.return_value": "behaviour_id_1"},
        )
        behaviour_2 = MagicMock(
            matching_round=MagicMock(),
            **{"auto_behaviour_id.return_value": "behaviour_id_2"},
        )

        with pytest.raises(
            ABCIAppInternalError,
            match=f"initial behaviour {behaviour_2.auto_behaviour_id()} is not in the set of behaviours",
        ):

            class MyRoundBehaviour(AbstractRoundBehaviour):
                abci_app_cls = ConcreteAbciApp
                behaviours = {behaviour_1}
                initial_behaviour_cls = behaviour_2

    def test_act_no_round_change(self) -> None:
        """Test the 'act' method of the behaviour, with no round change."""
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # check that after setup(), current behaviour is initial behaviour
        self.behaviour.setup()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        with mock.patch.object(
            self.behaviour.current_behaviour, "clean_up"
        ) as clean_up_mock:
            # check that after act(), current behaviour is initial behaviour and `clean_up()` has not been called
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourA)
            clean_up_mock.assert_not_called()

            # check that once the flag done is set, the `clean_up()` has been called
            # and `current_behaviour` is set to `None`.
            self.behaviour.current_behaviour.set_done()
            self.behaviour.act()
            assert self.behaviour.current_behaviour is None
            clean_up_mock.assert_called_once()

    def test_act_behaviour_setup(self) -> None:
        """Test the 'act' method of the FSM behaviour triggers setup() of the behaviour."""
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # check that after setup(), current behaviour is initial behaviour
        self.behaviour.setup()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        assert self.behaviour.current_behaviour.count == 0

        with mock.patch.object(
            self.behaviour.current_behaviour, "clean_up"
        ) as clean_up_mock:
            # check that after act() first time, a call to setup has been made
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourA)
            assert self.behaviour.current_behaviour.count == 1

            # check that after act() second time, no further call to setup
            self.behaviour.act()
            assert self.behaviour.current_behaviour.count == 1

            # check that the `clean_up()` has not been called
            clean_up_mock.assert_not_called()

    def test_act_with_round_change(self) -> None:
        """Test the 'act' method of the behaviour, with round change."""
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # check that after setup(), current behaviour is initial behaviour
        self.behaviour.setup()
        assert isinstance(self.behaviour.current_behaviour, BehaviourA)

        # check that after act(), current behaviour is initial behaviour
        with mock.patch.object(
            self.behaviour.current_behaviour, "clean_up"
        ) as clean_up_mock:
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourA)
            clean_up_mock.assert_not_called()

            # change the round
            self.round_sequence_mock.current_round = RoundB(MagicMock(), MagicMock())
            self.round_sequence_mock.current_round_height = (
                self.round_sequence_mock.current_round_height + 1
            )

            # check that if the round is changed, the behaviour transition is performed and the clean-up is called
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourB)
            clean_up_mock.assert_called_once()

    def test_act_with_round_change_after_current_behaviour_is_none(self) -> None:
        """Test the 'act' method of the behaviour, with round change, after cur behaviour is none."""
        self.round_sequence_mock.current_round = RoundA(MagicMock(), MagicMock())
        self.round_sequence_mock.current_round_height = 0

        # instantiate behaviour
        self.behaviour.current_behaviour = self.behaviour.instantiate_behaviour_cls(BehaviourA)  # type: ignore
        self.behaviour.background_behaviour = self.behaviour.instantiate_behaviour_cls(ConcreteBackgroundBehaviour)  # type: ignore

        with mock.patch.object(
            self.behaviour.current_behaviour, "clean_up"
        ) as clean_up_mock:
            # check that after act(), current behaviour is same behaviour
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourA)
            clean_up_mock.assert_not_called()

            # check that after the behaviour is done, current behaviour is None
            self.behaviour.current_behaviour.set_done()
            self.behaviour.act()
            assert self.behaviour.current_behaviour is None
            clean_up_mock.assert_called_once()

            # change the round
            self.round_sequence_mock.current_round = RoundB(MagicMock(), MagicMock())
            self.round_sequence_mock.current_round_height = (
                self.round_sequence_mock.current_round_height + 1
            )

            # check that if the round is changed, the behaviour transition is taken
            self.behaviour.act()
            assert isinstance(self.behaviour.current_behaviour, BehaviourB)
            clean_up_mock.assert_called_once()

    @mock.patch.object(
        AbstractRoundBehaviour,
        "_process_current_round",
    )
    @mock.patch.object(
        TmManager,
        "tm_communication_unhealthy",
        new_callable=mock.PropertyMock,
        return_value=False,
    )
    @mock.patch.object(
        TmManager,
        "is_acting",
        new_callable=mock.PropertyMock,
        return_value=False,
    )
    @pytest.mark.parametrize("expected_background_acting", (True, False))
    def test_background_behaviour_acting(
        self,
        _: mock._patch,
        __: mock._patch,
        ___: mock._patch,
        expected_background_acting: bool,
    ) -> None:
        """Test if the background behaviour is acting only when it should."""
        self.behaviour.context.params.use_termination = expected_background_acting
        self.behaviour.setup()
        if expected_background_acting:
            with mock.patch.object(
                self.behaviour.background_behaviour,
                "act_wrapper",
            ) as mock_background_act:
                self.behaviour.act()
                mock_background_act.assert_called()
        else:
            assert self.behaviour.background_behaviour is None

    @mock.patch.object(
        AbstractRoundBehaviour,
        "_process_current_round",
    )
    @pytest.mark.parametrize(
        ("mock_tm_communication_unhealthy", "mock_is_acting", "expected_fix"),
        [
            (True, True, True),
            (False, True, True),
            (True, False, True),
            (False, False, False),
        ],
    )
    def test_try_fix_call(
        self,
        _: mock._patch,
        mock_tm_communication_unhealthy: bool,
        mock_is_acting: bool,
        expected_fix: bool,
    ) -> None:
        """Test that `try_fix` is called when necessary."""
        with mock.patch.object(
            TmManager,
            "tm_communication_unhealthy",
            new_callable=mock.PropertyMock,
            return_value=mock_tm_communication_unhealthy,
        ), mock.patch.object(
            TmManager,
            "is_acting",
            new_callable=mock.PropertyMock,
            return_value=mock_is_acting,
        ), mock.patch.object(
            TmManager,
            "try_fix",
        ) as mock_try_fix:
            self.behaviour.act()
            if expected_fix:
                mock_try_fix.assert_called()
            else:
                mock_try_fix.assert_not_called()


def test_meta_round_behaviour_when_instance_not_subclass_of_abstract_round_behaviour() -> (
    None
):
    """Test instantiation of meta class when instance not a subclass of abstract round behaviour."""

    class MyRoundBehaviour(metaclass=_MetaRoundBehaviour):
        pass


def test_abstract_round_behaviour_instantiation_without_attributes_raises_error() -> (
    None
):
    """Test that definition of concrete subclass of AbstractRoundBehavior without attributes raises error."""
    with pytest.raises(ABCIAppInternalError):

        class MyRoundBehaviour(AbstractRoundBehaviour):
            pass


def test_abstract_round_behaviour_matching_rounds_not_covered() -> None:
    """Test that definition of concrete subclass of AbstractRoundBehavior when matching round not covered."""
    with pytest.raises(ABCIAppInternalError):

        class MyRoundBehaviour(AbstractRoundBehaviour):
            abci_app_cls = ConcreteAbciApp
            behaviours = {BehaviourA}
            initial_behaviour_cls = BehaviourA


@mock.patch.object(
    BaseBehaviour,
    "tm_communication_unhealthy",
    new_callable=mock.PropertyMock,
    return_value=False,
)
def test_self_loops_in_abci_app_reinstantiate_behaviour(_: mock._patch) -> None:
    """Test that a self-loop transition in the AbciApp will trigger a transition in the round behaviour."""
    event = MagicMock()

    class AbciAppTest(AbciApp):
        initial_round_cls = RoundA
        transition_function = {RoundA: {event: RoundA}}

    class RoundBehaviour(AbstractRoundBehaviour):
        abci_app_cls = AbciAppTest
        behaviours = {BehaviourA}
        initial_behaviour_cls = BehaviourA

    round_sequence = RoundSequence(AbciAppTest)
    round_sequence.end_sync()
    round_sequence.setup(MagicMock(), MagicMock())
    context_mock = MagicMock()
    context_mock.state.round_sequence = round_sequence
    behaviour = RoundBehaviour(name="", skill_context=context_mock)
    behaviour.setup()

    behaviour_1 = behaviour.current_behaviour
    assert isinstance(behaviour_1, BehaviourA)

    round_sequence.abci_app.process_event(event)

    behaviour.act()
    behaviour_2 = behaviour.current_behaviour
    assert isinstance(behaviour_2, BehaviourA)
    assert id(behaviour_1) != id(behaviour_2)
    assert behaviour_1 != behaviour_2


class LongRunningBehaviour(BaseBehaviour):
    """A behaviour that runs forevever."""

    behaviour_id = "long_running_behaviour"
    matching_round = RoundA

    def async_act(self) -> Generator:
        """An act method that simply cycles forever."""
        while True:
            # cycle forever
            yield


def test_reset_should_be_performed_when_tm_unhealthy() -> None:
    """Test that hard reset is performed while a behaviour is running, and tendermint communication is unhealthy."""
    event = MagicMock()

    class AbciAppTest(AbciApp):
        initial_round_cls = RoundA
        transition_function = {RoundA: {event: RoundA}}

    class RoundBehaviour(AbstractRoundBehaviour):
        abci_app_cls = AbciAppTest
        behaviours = {LongRunningBehaviour}  # type: ignore
        initial_behaviour_cls = LongRunningBehaviour

    round_sequence = RoundSequence(AbciAppTest)
    round_sequence.end_sync()
    round_sequence.setup(MagicMock(), MagicMock())
    context_mock = MagicMock()
    context_mock.state.round_sequence = round_sequence
    tm_recovery_params = TendermintRecoveryParams(
        reset_from_round=RoundA.auto_round_id()
    )
    context_mock.state.get_acn_result = MagicMock(return_value=tm_recovery_params)
    context_mock.params.ipfs_domain_name = None
    behaviour = RoundBehaviour(name="", skill_context=context_mock)
    behaviour.setup()

    current_behaviour = behaviour.current_behaviour
    assert isinstance(current_behaviour, LongRunningBehaviour)

    # upon entering the behaviour, the tendermint node communication is working well
    with mock.patch.object(
        RoundSequence,
        "block_stall_deadline_expired",
        new_callable=mock.PropertyMock,
        return_value=False,
    ):
        behaviour.act()

    def dummy_num_peers(
        timeout: Optional[float] = None,
    ) -> Generator[None, None, Optional[int]]:
        """A dummy method for num_active_peers."""
        # a None response is acceptable here, because tendermint is not healthy
        return None
        yield

    def dummy_reset_tendermint_with_wait(
        on_startup: bool = False,
        is_recovery: bool = False,
    ) -> Generator[None, None, bool]:
        """A dummy method for reset_tendermint_with_wait."""
        # we assume the reset goes through successfully
        return True
        yield

    # at this point LongRunningBehaviour is running
    # while the behaviour is running, the tendermint node
    # becomes unhealthy, we expect the node to be reset
    with mock.patch.object(
        RoundSequence,
        "block_stall_deadline_expired",
        new_callable=mock.PropertyMock,
        return_value=True,
    ), mock.patch.object(
        BaseBehaviour,
        "num_active_peers",
        side_effect=dummy_num_peers,
    ), mock.patch.object(
        BaseBehaviour,
        "reset_tendermint_with_wait",
        side_effect=dummy_reset_tendermint_with_wait,
    ) as mock_reset_tendermint:
        assert behaviour.tm_manager is not None
        behaviour.tm_manager.gentle_reset_attempted = True
        behaviour.act()
        mock_reset_tendermint.assert_called()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `run` command."""


import os
from unittest import mock

from autonomy.constants import DOCKER_COMPOSE_YAML

from tests.test_autonomy.test_cli.base import BaseCliTest


class TestRun(BaseCliTest):
    """Test `run` command."""

    cli_options = ("deploy", "run")

    def setup(self) -> None:
        """Setup test method."""

        super().setup()
        os.chdir(self.t)

    def test_run(
        self,
    ) -> None:
        """Run test."""
        (self.t / DOCKER_COMPOSE_YAML).touch()
        with mock.patch(
            "autonomy.cli.helpers.deployment.docker_compose.project_from_options"
        ), mock.patch("autonomy.cli.helpers.deployment.docker_compose.TopLevelCommand"):
            result = self.run_cli()
            assert result.exit_code == 0, result.output
            assert "Running build @" in result.output

    def test_missing_config_file(
        self,
    ) -> None:
        """Run test."""
        result = self.run_cli()
        assert result.exit_code == 1, result.output
        assert "Deployment configuration does not exist" in result.output



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""On-chain tools configurations."""

import os
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Optional, cast

from autonomy.chain import constants as chain_constants


DEFAULT_LOCAL_RPC = "http://127.0.0.1:8545"
DEFAULT_LOCAL_CHAIN_ID = 31337
CUSTOM_CHAIN_RPC = "CUSTOM_CHAIN_RPC"
GOERLI_CHAIN_RPC = "GOERLI_CHAIN_RPC"
ETHEREUM_CHAIN_RPC = "ETHEREUM_CHAIN_RPC"


def _get_chain_id_for_custom_chain() -> Optional[int]:
    """Get chain if for custom chain from environment"""
    chain_id = os.environ.get("CUSTOM_CHAIN_ID")
    if chain_id is None:
        return None

    return int(chain_id)


class ChainType(Enum):
    """Chain types."""

    LOCAL = "local"
    CUSTOM = "custom_chain"
    GOERLI = "goerli"
    ETHEREUM = "ethereum"


@dataclass
class ContractConfig:
    """Contract config class."""

    name: str
    contracts: Dict[ChainType, str]


@dataclass
class ChainConfig:
    """Chain config"""

    chain_type: ChainType
    rpc: Optional[str]
    chain_id: Optional[int]


class ChainConfigs:  # pylint: disable=too-few-public-methods
    """Name space for chain configs."""

    local = ChainConfig(
        chain_type=ChainType.LOCAL,
        rpc=DEFAULT_LOCAL_RPC,
        chain_id=DEFAULT_LOCAL_CHAIN_ID,
    )

    custom_chain = ChainConfig(
        chain_type=ChainType.CUSTOM,
        rpc=os.environ.get(CUSTOM_CHAIN_RPC),
        chain_id=_get_chain_id_for_custom_chain(),
    )

    goerli = ChainConfig(
        chain_type=ChainType.GOERLI,
        rpc=os.environ.get(GOERLI_CHAIN_RPC),
        chain_id=5,
    )

    ethereum = ChainConfig(
        chain_type=ChainType.ETHEREUM,
        rpc=os.environ.get(ETHEREUM_CHAIN_RPC),
        chain_id=1,
    )

    _rpc_env_vars = {
        ChainType.CUSTOM: CUSTOM_CHAIN_RPC,
        ChainType.GOERLI: GOERLI_CHAIN_RPC,
        ChainType.ETHEREUM: ETHEREUM_CHAIN_RPC,
    }

    @classmethod
    def get(cls, chain_type: ChainType) -> ChainConfig:
        """Return chain config for given chain type."""

        return cast(ChainConfig, getattr(cls, chain_type.value))

    @classmethod
    def get_rpc_env_var(cls, chain_type: ChainType) -> Optional[str]:
        """Return chain config for given chain type."""

        return cls._rpc_env_vars.get(chain_type)


class ContractConfigs:  # pylint: disable=too-few-public-methods
    """A namespace for contract configs."""

    component_registry = ContractConfig(
        name="component_registry",
        contracts={
            chain_type: getattr(
                chain_constants, f"COMPONENT_REGISTRY_ADDRESS_{chain_type.name}"
            )
            for chain_type in ChainType
        },
    )

    agent_registry = ContractConfig(
        name="agent_registry",
        contracts={
            chain_type: getattr(
                chain_constants, f"AGENT_REGISTRY_ADDRESS_{chain_type.name}"
            )
            for chain_type in ChainType
        },
    )

    service_registry = ContractConfig(
        name="service_registry",
        contracts={
            chain_type: getattr(
                chain_constants, f"SERVICE_REGISTRY_ADDRESS_{chain_type.name}"
            )
            for chain_type in ChainType
        },
    )

    service_manager = ContractConfig(
        name="service_manager",
        contracts={
            chain_type: getattr(
                chain_constants, f"SERVICE_MANAGER_ADDRESS_{chain_type.name}"
            )
            for chain_type in ChainType
        },
    )

    registries_manager = ContractConfig(
        name="registries_manager",
        contracts={
            chain_type: getattr(
                chain_constants, f"REGISTRIES_MANAGER_ADDRESS_{chain_type.name}"
            )
            for chain_type in ChainType
        },
    )

    gnosis_safe_proxy_factory = ContractConfig(
        name="gnosis_safe_proxy_factory",
        contracts={
            chain_type: getattr(
                chain_constants, f"GNOSIS_SAFE_MULTISIG_ADDRESS_{chain_type.name}"
            )
            for chain_type in ChainType
        },
    )

    @classmethod
    def get(cls, name: str) -> ContractConfig:
        """Return chain config for given chain type."""

        return cast(ContractConfig, getattr(cls, name))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/hello_world_abci skill's behaviours."""

# pylint: skip-file

import json
import time
from pathlib import Path
from typing import Any, Type, cast
from unittest import mock

from aea.skills.base import SkillContext

from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.abstract_round_abci.behaviour_utils import BaseBehaviour
from packages.valory.skills.abstract_round_abci.test_tools.base import (
    FSMBehaviourBaseCase,
)
from packages.valory.skills.hello_world_abci import PUBLIC_ID
from packages.valory.skills.hello_world_abci.behaviours import (
    CollectRandomnessBehaviour,
    PrintMessageBehaviour,
    RegistrationBehaviour,
    ResetAndPauseBehaviour,
    SelectKeeperBehaviour,
)
from packages.valory.skills.hello_world_abci.rounds import Event, SynchronizedData


PACKAGE_DIR = Path(__file__).parent.parent


def test_skill_public_id() -> None:
    """Test skill module public ID"""

    assert PUBLIC_ID.name == Path(__file__).parents[1].name
    assert PUBLIC_ID.author == Path(__file__).parents[3].name


class HelloWorldAbciFSMBehaviourBaseCase(FSMBehaviourBaseCase):
    """Base case for testing PriceEstimation FSMBehaviour."""

    path_to_skill = PACKAGE_DIR

    def setup(self, **kwargs: Any) -> None:
        """
        Set up the test method.

        Called each time before a test method is called.

        :param kwargs: the keyword arguments passed to _prepare_skill
        """
        super().setup(**kwargs)
        self.synchronized_data = SynchronizedData(
            AbciAppDB(
                setup_data=dict(
                    most_voted_keeper_address=["most_voted_keeper_address"],
                ),
            )
        )

    def end_round(  # type: ignore
        self,
    ) -> None:
        """Ends round early to cover `wait_for_end` generator."""
        super().end_round(Event.DONE)


class BaseCollectRandomnessBehaviourTest(HelloWorldAbciFSMBehaviourBaseCase):
    """Test CollectRandomnessBehaviour."""

    collect_randomness_behaviour_class: Type[BaseBehaviour]
    next_behaviour_class: Type[BaseBehaviour]

    def test_randomness_behaviour(
        self,
    ) -> None:
        """Test RandomnessBehaviour."""

        self.fast_forward_to_behaviour(
            self.behaviour,
            self.collect_randomness_behaviour_class.auto_behaviour_id(),
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.collect_randomness_behaviour_class.auto_behaviour_id()
        )
        self.behaviour.act_wrapper()
        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                headers="",
                version="",
                body=b"",
                url="https://drand.cloudflare.com/public/latest",
            ),
            response_kwargs=dict(
                version="",
                status_code=200,
                status_text="",
                headers="",
                body=json.dumps(
                    {
                        "round": 1283255,
                        "randomness": "04d4866c26e03347d2431caa82ab2d7b7bdbec8b58bca9460c96f5265d878feb",
                    }
                ).encode("utf-8"),
            ),
        )

        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()

        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.auto_behaviour_id()

    def test_invalid_response(
        self,
    ) -> None:
        """Test invalid json response."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.collect_randomness_behaviour_class.auto_behaviour_id(),
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.collect_randomness_behaviour_class.auto_behaviour_id()
        )
        self.behaviour.act_wrapper()

        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                headers="",
                version="",
                body=b"",
                url="https://drand.cloudflare.com/public/latest",
            ),
            response_kwargs=dict(
                version="", status_code=200, status_text="", headers="", body=b""
            ),
        )
        self.behaviour.act_wrapper()
        time.sleep(1)
        self.behaviour.act_wrapper()

    def test_max_retries_reached(
        self,
    ) -> None:
        """Test with max retries reached."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.collect_randomness_behaviour_class.auto_behaviour_id(),
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.collect_randomness_behaviour_class.auto_behaviour_id()
        )
        self.behaviour.context.randomness_api.__dict__["_frozen"] = False
        with mock.patch.object(
            self.behaviour.context.randomness_api,
            "is_retries_exceeded",
            return_value=True,
        ):
            self.behaviour.act_wrapper()
            behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
            assert (
                behaviour.behaviour_id
                == self.collect_randomness_behaviour_class.auto_behaviour_id()
            )
            self._test_done_flag_set()
        self.behaviour.context.randomness_api.__dict__["_frozen"] = True

    def test_clean_up(
        self,
    ) -> None:
        """Test when `observed` value is none."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.collect_randomness_behaviour_class.auto_behaviour_id(),
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.collect_randomness_behaviour_class.auto_behaviour_id()
        )
        self.behaviour.context.randomness_api.retries_info.retries_attempted = 1
        assert self.behaviour.current_behaviour is not None
        self.behaviour.current_behaviour.clean_up()
        assert self.behaviour.context.randomness_api.retries_info.retries_attempted == 0


class BaseSelectKeeperBehaviourTest(HelloWorldAbciFSMBehaviourBaseCase):
    """Test SelectKeeperBehaviour."""

    select_keeper_behaviour_class: Type[BaseBehaviour]
    next_behaviour_class: Type[BaseBehaviour]

    def test_select_keeper(
        self,
    ) -> None:
        """Test select keeper agent."""
        participants = (self.skill.skill_context.agent_address, "a_1", "a_2")
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.select_keeper_behaviour_class.auto_behaviour_id(),
            synchronized_data=SynchronizedData(
                AbciAppDB(
                    setup_data=dict(
                        participants=[participants],
                        most_voted_randomness=[
                            "56cbde9e9bbcbdcaf92f183c678eaa5288581f06b1c9c7f884ce911776727688"
                        ],
                        most_voted_keeper_address=["most_voted_keeper_address"],
                    ),
                )
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.select_keeper_behaviour_class.auto_behaviour_id()
        )
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.auto_behaviour_id()


class TestRegistrationBehaviour(HelloWorldAbciFSMBehaviourBaseCase):
    """Test case to test RegistrationBehaviour."""

    def test_registration(self) -> None:
        """Test registration."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            RegistrationBehaviour.auto_behaviour_id(),
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == RegistrationBehaviour.auto_behaviour_id()
        )
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()

        self.end_round()
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == CollectRandomnessBehaviour.auto_behaviour_id()


class TestCollectRandomnessBehaviour(BaseCollectRandomnessBehaviourTest):
    """Test CollectRandomnessBehaviour."""

    collect_randomness_behaviour_class = CollectRandomnessBehaviour
    next_behaviour_class = SelectKeeperBehaviour


class TestSelectKeeperBehaviour(BaseSelectKeeperBehaviourTest):
    """Test SelectKeeperBehaviour."""

    select_keeper_behaviour_class = SelectKeeperBehaviour
    next_behaviour_class = PrintMessageBehaviour


class TestPrintMessageBehaviour(HelloWorldAbciFSMBehaviourBaseCase):
    """Test case to test PrintMessageBehaviour."""

    def test_print_message_non_keeper(self) -> None:
        """Test print_message."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            PrintMessageBehaviour.auto_behaviour_id(),
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == PrintMessageBehaviour.auto_behaviour_id()
        )
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()

        self.end_round()
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == ResetAndPauseBehaviour.auto_behaviour_id()

    @mock.patch.object(SkillContext, "agent_address", new_callable=mock.PropertyMock)
    def test_print_message_keeper(
        self,
        agent_address_mock: mock.PropertyMock,
    ) -> None:
        """Test print_message."""
        agent_address_mock.return_value = "most_voted_keeper_address"
        self.fast_forward_to_behaviour(
            self.behaviour,
            PrintMessageBehaviour.auto_behaviour_id(),
            self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == PrintMessageBehaviour.auto_behaviour_id()
        )
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()

        self.end_round()
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == ResetAndPauseBehaviour.auto_behaviour_id()


class TestResetAndPauseBehaviour(HelloWorldAbciFSMBehaviourBaseCase):
    """Test ResetBehaviour."""

    behaviour_class = ResetAndPauseBehaviour
    next_behaviour_class = CollectRandomnessBehaviour

    def test_pause_and_reset_behaviour(
        self,
    ) -> None:
        """Test pause and reset behaviour."""
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.behaviour_class.auto_behaviour_id(),
            synchronized_data=self.synchronized_data,
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.behaviour_class.auto_behaviour_id()
        )
        self.behaviour.context.params.__dict__["reset_pause_duration"] = 0.1
        self.behaviour.act_wrapper()
        time.sleep(0.3)
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.auto_behaviour_id()

    def test_reset_behaviour(
        self,
    ) -> None:
        """Test reset behaviour."""
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.behaviour_class.auto_behaviour_id(),
            synchronized_data=self.synchronized_data,
        )
        assert self.behaviour.current_behaviour is not None
        self.behaviour.current_behaviour.pause = False
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.behaviour_class.auto_behaviour_id()
        )
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round()
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.auto_behaviour_id()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Deployment helpers."""
import os
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import click
from aea.configurations.data_types import PublicId
from aea.helpers.base import cd
from compose.cli import main as docker_compose
from compose.config.errors import ConfigurationError
from docker.errors import NotFound

from autonomy.chain.config import ChainType, ContractConfigs
from autonomy.chain.exceptions import FailedToRetrieveComponentMetadata
from autonomy.chain.service import get_agent_instances, get_service_info
from autonomy.chain.utils import resolve_component_id
from autonomy.cli.helpers.chain import get_ledger_and_crypto_objects
from autonomy.cli.helpers.registry import fetch_service_ipfs
from autonomy.configurations.constants import DEFAULT_SERVICE_CONFIG_FILE
from autonomy.configurations.loader import load_service_config
from autonomy.constants import DEFAULT_BUILD_FOLDER
from autonomy.deploy.build import generate_deployment
from autonomy.deploy.constants import (
    AGENT_KEYS_DIR,
    BENCHMARKS_DIR,
    INFO,
    LOG_DIR,
    PERSISTENT_DATA_DIR,
    TM_STATE_DIR,
    VENVS_DIR,
)
from autonomy.deploy.generators.kubernetes.base import KubernetesGenerator
from autonomy.deploy.image import build_image


def _build_dirs(build_dir: Path) -> None:
    """Build necessary directories."""

    for dir_path in [
        (PERSISTENT_DATA_DIR,),
        (PERSISTENT_DATA_DIR, LOG_DIR),
        (PERSISTENT_DATA_DIR, TM_STATE_DIR),
        (PERSISTENT_DATA_DIR, BENCHMARKS_DIR),
        (PERSISTENT_DATA_DIR, VENVS_DIR),
        (AGENT_KEYS_DIR,),
    ]:
        path = Path(build_dir, *dir_path)
        path.mkdir()
        # TOFIX: remove this safely
        try:
            os.chown(path, 1000, 1000)
        except PermissionError:  # pragma: no cover
            click.echo(
                f"Updating permissions failed for {path}, please do it manually."
            )


def run_deployment(
    build_dir: Path, no_recreate: bool = False, remove_orphans: bool = False
) -> None:
    """Run deployment."""

    click.echo(f"Running build @ {build_dir}")
    try:
        project = docker_compose.project_from_options(build_dir, {})
    except ConfigurationError as e:  # pragma: no cover
        if "Invalid interpolation format" in e.msg:
            raise click.ClickException(
                "Provided docker compose file contains environment placeholders, "
                "please use `--aev` flag if you intend to use environment variables."
            )
        raise

    try:
        commands = docker_compose.TopLevelCommand(project=project)
        commands.up(
            {
                "--detach": False,
                "--no-color": False,
                "--quiet-pull": False,
                "--no-deps": False,
                "--force-recreate": not no_recreate,
                "--always-recreate-deps": False,
                "--no-recreate": no_recreate,
                "--no-build": False,
                "--no-start": False,
                "--build": True,
                "--abort-on-container-exit": False,
                "--attach-dependencies": False,
                "--timeout": None,
                "--renew-anon-volumes": False,
                "--remove-orphans": remove_orphans,
                "--exit-code-from": None,
                "--scale": [],
                "--no-log-prefix": False,
                "SERVICE": None,
            }
        )
    except NotFound as e:  # pragma: no cover
        raise click.ClickException(e.explanation)


def build_deployment(  # pylint: disable=too-many-arguments, too-many-locals
    keys_file: Path,
    build_dir: Path,
    deployment_type: str,
    dev_mode: bool,
    number_of_agents: Optional[int] = None,
    password: Optional[str] = None,
    packages_dir: Optional[Path] = None,
    open_aea_dir: Optional[Path] = None,
    open_autonomy_dir: Optional[Path] = None,
    agent_instances: Optional[List[str]] = None,
    multisig_address: Optional[str] = None,
    consensus_threshold: Optional[int] = None,
    log_level: str = INFO,
    apply_environment_variables: bool = False,
    image_version: Optional[str] = None,
    use_hardhat: bool = False,
    use_acn: bool = False,
    use_tm_testnet_setup: bool = False,
    image_author: Optional[str] = None,
) -> None:
    """Build deployment."""

    if build_dir.is_dir():  # pragma: no cover
        raise click.ClickException(f"Build already exists @ {build_dir}")

    if not (Path.cwd() / DEFAULT_SERVICE_CONFIG_FILE).exists():
        raise FileNotFoundError(
            f"No service configuration found at {Path.cwd()}"
        )  # pragma: no cover

    click.echo(f"Building deployment @ {build_dir}")
    build_dir.mkdir()
    _build_dirs(build_dir)

    report = generate_deployment(
        service_path=Path.cwd(),
        type_of_deployment=deployment_type,
        keys_file=keys_file,
        private_keys_password=password,
        number_of_agents=number_of_agents,
        build_dir=build_dir,
        dev_mode=dev_mode,
        packages_dir=packages_dir,
        open_aea_dir=open_aea_dir,
        open_autonomy_dir=open_autonomy_dir,
        agent_instances=agent_instances,
        multisig_address=multisig_address,
        consensus_threshold=consensus_threshold,
        log_level=log_level,
        apply_environment_variables=apply_environment_variables,
        image_version=image_version,
        use_hardhat=use_hardhat,
        use_acn=use_acn,
        use_tm_testnet_setup=use_tm_testnet_setup,
        image_author=image_author,
    )
    click.echo(report)


def _resolve_on_chain_token_id(
    token_id: int,
    chain_type: ChainType,
) -> Tuple[Dict[str, str], List[str], str, int]:
    """Resolve service metadata from tokenID"""

    ledger_api, _ = get_ledger_and_crypto_objects(chain_type=chain_type)
    contract_address = ContractConfigs.service_registry.contracts[chain_type]

    click.echo(f"Fetching service metadata using chain type {chain_type.value}")

    try:
        metadata = resolve_component_id(
            ledger_api=ledger_api, contract_address=contract_address, token_id=token_id
        )
        info = get_agent_instances(
            ledger_api=ledger_api, chain_type=chain_type, token_id=token_id
        )
        agent_instances = info["agentInstances"]
        (
            _,
            multisig_address,
            _,
            consensus_threshold,
            *_,
        ) = get_service_info(
            ledger_api=ledger_api, chain_type=chain_type, token_id=token_id
        )
    except FailedToRetrieveComponentMetadata as e:
        raise click.ClickException(str(e)) from e
    except Exception as e:
        raise click.ClickException(
            f"Cannot find the service registry deployment; Service contract address {contract_address}; Error: {e}"
        ) from e

    return metadata, agent_instances, multisig_address, consensus_threshold


def build_and_deploy_from_token(  # pylint: disable=too-many-arguments, too-many-locals
    token_id: int,
    keys_file: Path,
    chain_type: ChainType,
    skip_image: bool,
    n: Optional[int],
    deployment_type: str,
    aev: bool = False,
    password: Optional[str] = None,
    no_deploy: bool = False,
) -> None:
    """Build and run deployment from tokenID."""

    click.echo(f"Building service deployment using token ID: {token_id}")
    (
        service_metadata,
        agent_instances,
        multisig_address,
        consensus_threshold,
    ) = _resolve_on_chain_token_id(
        token_id=token_id,
        chain_type=chain_type,
    )

    click.echo("Service name: " + service_metadata["name"])
    *_, service_hash = service_metadata["code_uri"].split("//")
    public_id = PublicId(author="valory", name="service", package_hash=service_hash)
    service_path = fetch_service_ipfs(public_id)
    build_dir = service_path / DEFAULT_BUILD_FOLDER

    with cd(service_path):
        build_deployment(
            keys_file=keys_file,
            build_dir=build_dir,
            deployment_type=deployment_type,
            dev_mode=False,
            number_of_agents=n,
            agent_instances=agent_instances,
            multisig_address=multisig_address,
            consensus_threshold=consensus_threshold,
            apply_environment_variables=aev,
            password=password,
        )
        if not skip_image:
            click.echo("Building required images.")
            service = load_service_config(service_path, substitute_env_vars=aev)
            build_image(agent=service.agent)

    click.echo("Service build successful.")
    if no_deploy or deployment_type == KubernetesGenerator.deployment_type:
        return

    click.echo("Running deployment")
    run_deployment(build_dir)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `autonomy analyse dialgues` module."""

from pathlib import Path
from typing import List, Optional, Tuple
from unittest import mock

import yaml
from aea.configurations.constants import DEFAULT_SKILL_CONFIG_FILE

from autonomy.analyse.constants import CLASS_NAME, DIALOGUES_FILE

from tests.test_autonomy.test_cli.base import BaseCliTest


class BaseAnalyseDialoguesTest(BaseCliTest):
    """Base class for testing dialogues analyser"""

    cli_options: Tuple[str, ...] = ("analyse", "dialogues")
    skill_dir: Path

    def patch_list_skills(self) -> mock._patch:
        """Patch `autonomy.cli.helpers.analyse.list_all_skill_yaml_files` method to return temporary skill dir"""

        return mock.patch(
            "autonomy.cli.helpers.analyse.list_all_skill_yaml_files",
            return_value=[self.skill_dir / DEFAULT_SKILL_CONFIG_FILE],
        )

    def make_dummy_skill(
        self,
        config_data: List[Tuple[str, str]],
        dialogues: List[str],
    ) -> None:
        """Make duummy skill."""
        self.skill_dir = self.t / "some_skill"
        self.skill_dir.mkdir()

        config_file = self.skill_dir / DEFAULT_SKILL_CONFIG_FILE
        with open(config_file, "w") as fp:
            yaml.safe_dump(
                data={
                    "models": {
                        dialogue_name: {CLASS_NAME: dialogue_class}
                        for dialogue_name, dialogue_class in config_data
                    }
                },
                stream=fp,
            )

        if len(dialogues) > 0:
            dialogues_file = self.skill_dir / DIALOGUES_FILE
            dialogues_file.write_text(
                "\n".join([f"{dialogue_class} = None" for dialogue_class in dialogues])
            )


class TestAnalyseDialogues(BaseAnalyseDialoguesTest):
    """Test analyse dialogues."""

    def run_test(
        self,
        config_data: List[Tuple[str, str]],
        dialogues: List[str],
        exit_code: int,
        stdout_checks: List[str],
        commands: Optional[Tuple[str, ...]] = None,
    ) -> None:
        """Run test."""
        self.make_dummy_skill(
            config_data=config_data,
            dialogues=dialogues,
        )

        with self.patch_list_skills():
            result = self.run_cli(commands=(commands or ()))

        assert result.exit_code == exit_code, result.stdout

        for check in stdout_checks:
            assert check in result.stdout

    def test_analyse_default(self) -> None:
        """Test analyse."""

        self.run_test(
            config_data=[
                ("abci_dialogues", "AbciDialogues"),
            ],
            dialogues=["AbciDialogues"],
            exit_code=0,
            stdout_checks=["Checking"],
        )

    def test_analyse_fail_dialogues_does_not_exist(self) -> None:
        """Test analyse."""

        self.run_test(
            config_data=[],
            dialogues=[],
            exit_code=1,
            stdout_checks=["dialogue file does not exist"],
        )

    def test_analyse_fail_wrong_class_name(self) -> None:
        """Test analyse."""

        self.run_test(
            config_data=[
                ("abci_dialogues", "AbciDialogue"),
            ],
            dialogues=["AbciDialogue"],
            exit_code=1,
            stdout_checks=[
                "Class name for a dialogue should end with `Dialogues`; dialogue=abci_dialogues; class=AbciDialogue"
            ],
        )

    def test_analyse_fail_wrong_dialogue_name(self) -> None:
        """Test analyse."""

        self.run_test(
            config_data=[
                ("abci_dialogue", "AbciDialogues"),
            ],
            dialogues=["AbciDialogues"],
            exit_code=1,
            stdout_checks=[
                "Dialogue name should end with `dialogues`; dialogue=abci_dialogue; class=AbciDialogues"
            ],
        )

    def test_analyse_fail_common_dialogue_not_found(self) -> None:
        """Test analyse."""

        self.run_test(
            config_data=[
                ("abci_dialogues", "AbciDialogues"),
            ],
            dialogues=["AbciDialogues"],
            exit_code=1,
            stdout_checks=["Common dialogue 'some_dialogues' is not defined"],
            commands=("-d", "some_dialogues"),
        )

    def test_analyse_fail_common_dialogue_class_not_found(self) -> None:
        """Test analyse."""

        self.run_test(
            config_data=[
                ("abci_dialogues", "AbciDialogues"),
                ("some_dialogues", "SomeDialogues"),
            ],
            dialogues=["AbciDialogues"],
            exit_code=1,
            stdout_checks=["dialogue SomeDialogues declared in", "is missing from"],
            commands=("-d", "some_dialogues"),
        )

    def test_analyse_skip_file(self) -> None:
        """Test analyse."""

        self.run_test(
            config_data=[],
            dialogues=[],
            exit_code=0,
            stdout_checks=["Skipping some_skill"],
            commands=("-i", "some_skill"),
        )
